import logging
from collections.abc import Iterable
from typing import Optional, Type

# Django
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import connection, models, transaction
from django.db.models.functions import Cast
from django.db.models.query import QuerySet
from django.db.utils import IntegrityError
from django.utils.translation import gettext_lazy as _

# Django-rest-framework
from rest_framework.exceptions import ValidationError

# ansible_base lib functions
from ansible_base.lib.abstract_models.common import CommonModel, ImmutableCommonModel

# ansible_base RBAC logic imports
from ansible_base.lib.utils.models import is_add_perm
from ansible_base.rbac.permission_registry import permission_registry
from ansible_base.rbac.prefetch import TypesPrefetch
from ansible_base.rbac.validators import validate_assignment, validate_permissions_for_model

logger = logging.getLogger('ansible_base.rbac.models')


class DABPermission(models.Model):
    "This is a minimal copy of auth.Permission for internal use"

    name = models.CharField("name", max_length=255, help_text=_("The name of this permission."))
    content_type = models.ForeignKey(ContentType, models.CASCADE, verbose_name="content type", help_text=_("The content type this permission will apply to."))
    codename = models.CharField(
        "codename",
        max_length=100,
        help_text=_(
            "".join(
                [
                    "A codename for the permission, in the format {action}_{model_name}. ",
                    "Where action is typically the view set action (view/list/etc) from Django rest framework.",
                ]
            )
        ),
    )

    class Meta:
        app_label = 'dab_rbac'
        verbose_name = "permission"
        verbose_name_plural = "permissions"
        unique_together = [["content_type", "codename"]]
        ordering = ["content_type__model", "codename"]

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.codename}>"


class ManagedRoleFromSetting:
    def __init__(self, role_name):
        super().__init__()
        self.role_name = role_name

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.role_name not in obj._cache:
            obj._cache[self.role_name] = RoleDefinition.objects.filter(name=self.role_name).first()
        return obj._cache[self.role_name]


class ManagedRoleManager:
    def __init__(self, apps):
        self._cache = {}
        self.apps = apps

    def clear(self) -> None:
        "Clear any managed roles already loaded into the cache"
        self._cache = {}

    def __getattr__(self, attr):
        if attr in self._cache:
            return self._cache[attr]
        constructor = permission_registry.get_managed_role_constructor(attr)
        if constructor:
            rd, _ = constructor.get_or_create(self.apps)
            return rd


class RoleDefinitionManager(models.Manager):
    def contribute_to_class(self, cls: Type[models.Model], name: str) -> None:
        """After Django populates the model for the manager, attach the manager role manager"""
        super().contribute_to_class(cls, name)
        self.managed = ManagedRoleManager(self.model._meta.apps)

    def give_creator_permissions(self, user, obj) -> Optional['RoleUserAssignment']:
        # If the user is a superuser, no need to bother giving the creator permissions
        for super_flag in settings.ANSIBLE_BASE_BYPASS_SUPERUSER_FLAGS:
            if getattr(user, super_flag):
                return

        needed_actions = settings.ANSIBLE_BASE_CREATOR_DEFAULTS

        # User should get permissions to the object and any child objects under it
        model_and_children = set(cls for rel, cls in permission_registry.get_child_models(obj))
        model_and_children.add(type(obj))
        cts = ContentType.objects.get_for_models(*model_and_children).values()

        needed_perms = set()
        for perm in DABPermission.objects.filter(content_type__in=cts).prefetch_related('content_type'):
            action = perm.codename.split('_', 1)[0]
            if action in needed_actions:
                # do not save add permission on the object level, which does not make sense
                if is_add_perm(perm.codename) and perm.content_type.model == obj._meta.model_name:
                    continue
                needed_perms.add(perm.codename)

        has_permissions = set(RoleEvaluation.get_permissions(user, obj))
        has_permissions.update(user.singleton_permissions())
        if set(needed_perms) - set(has_permissions):
            kwargs = {'permissions': needed_perms, 'name': settings.ANSIBLE_BASE_ROLE_CREATOR_NAME.format(obj=obj, cls=type(obj))}
            defaults = {'content_type': ContentType.objects.get_for_model(obj)}
            try:
                rd, _ = self.get_or_create(defaults=defaults, **kwargs)
            except ValidationError:
                logger.warning(f'Creating role definition {kwargs["name"]} as manged role because this is not allow as a custom role')
                defaults['managed'] = True
                rd, _ = self.get_or_create(defaults=defaults, **kwargs)

            return rd.give_permission(user, obj)

    def get_or_create(self, permissions=(), defaults=None, **kwargs):
        "Add extra feature on top of existing get_or_create to use permissions list"
        if permissions:
            permissions = set(permissions)
            for existing_rd in self.prefetch_related('permissions'):
                existing_set = set(perm.codename for perm in existing_rd.permissions.all())
                if existing_set == permissions:
                    return (existing_rd, False)
            create_kwargs = kwargs.copy()
            if defaults:
                create_kwargs.update(defaults)
            return (self.create_from_permissions(permissions=permissions, **create_kwargs), True)
        return super().get_or_create(defaults=defaults, **kwargs)

    def create_from_permissions(self, permissions=(), **kwargs):
        "Create from a list of text-type permissions and do validation"
        perm_list = [permission_registry.permission_qs.get(codename=str_perm) for str_perm in permissions]

        ct = kwargs.get('content_type', None)
        if kwargs.get('content_type_id', None):
            ct = ContentType.objects.get(id=kwargs['content_type_id'])

        validate_permissions_for_model(perm_list, ct, managed=kwargs.get('managed', False))

        rd = self.create(**kwargs)
        rd.permissions.add(*perm_list)
        return rd


class RoleDefinition(CommonModel):
    "Abstract definition of the permissions a role will grant before it is associated to an object"

    class Meta:
        app_label = 'dab_rbac'
        ordering = ['id']
        verbose_name_plural = _('role_definition')

    name = models.TextField(db_index=True, unique=True, help_text=_("The name of this role."))
    description = models.TextField(blank=True, help_text=_("A description of this role."))
    managed = models.BooleanField(
        default=False, editable=False, help_text=_("Is this role managed by the system (not changeable by the users).")
    )  # pulp definition of Role uses locked
    permissions = models.ManyToManyField('dab_rbac.DABPermission', related_name='role_definitions')
    content_type = models.ForeignKey(
        ContentType,
        help_text=_('The type of resource this can apply to; only used for validation and user assistance.'),
        null=True,
        default=None,
        on_delete=models.CASCADE,
    )

    objects = RoleDefinitionManager()
    router_basename = 'roledefinition'
    ignore_relations = ['permissions', 'object_roles', 'content_type', 'teams', 'users']

    def __str__(self):
        managed_str = ''
        if self.managed:
            managed_str = ', managed=True'
        return f'RoleDefinition(pk={self.id}, name={self.name}{managed_str})'

    def give_global_permission(self, actor):
        return self.give_or_remove_global_permission(actor, giving=True)

    def remove_global_permission(self, actor):
        return self.give_or_remove_global_permission(actor, giving=False)

    def give_or_remove_global_permission(self, actor, giving=True):
        if giving and (self.content_type is not None):
            raise ValidationError('Role definition content type must be null to assign globally')

        if actor._meta.model_name == 'user':
            if giving and (not settings.ANSIBLE_BASE_ALLOW_SINGLETON_USER_ROLES):
                raise ValidationError('Global roles are not enabled for users')
            kwargs = dict(object_role=None, user=actor, role_definition=self)
            cls = RoleUserAssignment
        elif isinstance(actor, permission_registry.team_model):
            if not settings.ANSIBLE_BASE_ALLOW_SINGLETON_TEAM_ROLES:
                raise ValidationError('Global roles are not enabled for teams')
            kwargs = dict(object_role=None, team=actor, role_definition=self)
            cls = RoleTeamAssignment
        else:
            raise RuntimeError(f'Cannot {giving and "give" or "remove"} permission for {actor}, must be a user or team')

        if giving:
            assignment, _ = cls.objects.get_or_create(**kwargs)
        else:
            assignment = cls.objects.filter(**kwargs).first()
            if assignment:
                assignment.delete()

        # Clear any cached permissions
        if actor._meta.model_name == 'user':
            if hasattr(actor, '_singleton_permissions'):
                delattr(actor, '_singleton_permissions')
        else:
            # when team permissions change, users in memory may be affected by this
            # but there is no way to know what users, so we use a global flag
            from ansible_base.rbac.evaluations import bound_singleton_permissions

            bound_singleton_permissions._team_clear_signal = True

        return assignment

    def give_permission(self, actor, content_object):
        return self.give_or_remove_permission(actor, content_object, giving=True)

    def remove_permission(self, actor, content_object):
        return self.give_or_remove_permission(actor, content_object, giving=False)

    def get_or_create_object_role(self, **kwargs):
        """Transaction-safe method to create ObjectRole

        The UI will assign many permissions concurrently.
        These will be in transactions, but also mutually create the same ObjectRole
        postgres constraints will still be violated by other active transactions
        which gives us a way to gracefully handle this.
        """
        if transaction.get_connection().in_atomic_block:
            try:
                with transaction.atomic():
                    object_role = ObjectRole.objects.create(**kwargs)
                    return (object_role, True)
            except IntegrityError:
                object_role = ObjectRole.objects.get(**kwargs)
                return (object_role, False)
        else:
            object_role = ObjectRole.objects.create(**kwargs)
            return (object_role, True)

    def give_or_remove_permission(self, actor, content_object, giving=True, sync_action=False):
        "Shortcut method to do whatever needed to give user or team these permissions"
        validate_assignment(self, actor, content_object)
        obj_ct = ContentType.objects.get_for_model(content_object)
        # sanitize the object_id to its database version, practically, remove "-" chars from uuids
        object_id = content_object._meta.pk.get_db_prep_value(content_object.pk, connection)
        kwargs = dict(role_definition=self, content_type=obj_ct, object_id=object_id)

        created = False
        object_role = ObjectRole.objects.filter(**kwargs).first()
        if object_role is None:
            if not giving:
                return  # nothing to do
            object_role, created = self.get_or_create_object_role(**kwargs)

        from ansible_base.rbac.triggers import needed_updates_on_assignment, update_after_assignment

        update_teams, to_update = needed_updates_on_assignment(self, actor, object_role, created=created, giving=True)

        assignment = None
        if actor._meta.model_name == 'user':
            if giving:
                assignment, created = RoleUserAssignment.objects.get_or_create(user=actor, object_role=object_role)
            else:
                object_role.users.remove(actor)
        elif isinstance(actor, permission_registry.team_model):
            if giving:
                assignment, created = RoleTeamAssignment.objects.get_or_create(team=actor, object_role=object_role)
            else:
                object_role.teams.remove(actor)

        if (not giving) and (not (object_role.users.exists() or object_role.teams.exists())):
            if object_role in to_update:
                to_update.remove(object_role)
            object_role.delete()

        update_after_assignment(update_teams, to_update)

        if not sync_action and self.name in permission_registry._trackers:
            tracker = permission_registry._trackers[self.name]
            with tracker.sync_active():
                tracker.sync_relationship(actor, content_object, giving=giving)

        return assignment

    @classmethod
    def user_global_permissions(cls, user, permission_qs=None):
        """Evaluation method only for global permissions from global roles

        This is special, in that it bypasses the RoleEvaluation table and methods.
        That is because global roles do not enumerate role permissions there,
        so global permissions are computed separately, here.
        """
        if permission_qs is None:
            # Allowing caller to replace the base permission set allows changing the type of thing returned
            # this is used in the assignment querysets, but these cases must call the method directly
            permission_qs = DABPermission.objects.all()

        perm_set = set()
        if settings.ANSIBLE_BASE_ALLOW_SINGLETON_USER_ROLES:
            rd_qs = cls.objects.filter(user_assignments__user=user, content_type=None)
            perm_qs = permission_qs.filter(role_definitions__in=rd_qs)
            perm_set.update(perm_qs)
        if settings.ANSIBLE_BASE_ALLOW_SINGLETON_TEAM_ROLES:
            # Users gain team membership via object roles that grant the teams member permission
            user_obj_roles = ObjectRole.objects.filter(users=user)
            user_teams_qs = permission_registry.team_model.objects.filter(member_roles__in=user_obj_roles)
            # Those teams (the user is in) then have a set of global roles they have been assigned
            rd_qs = cls.objects.filter(team_assignments__team__in=user_teams_qs, content_type=None)
            perm_qs = permission_qs.filter(role_definitions__in=rd_qs)
            perm_set.update(perm_qs)
        return perm_set

    def summary_fields(self):
        return {'id': self.id, 'name': self.name, 'description': self.description, 'managed': self.managed}


class ObjectRoleFields(models.Model):
    "Fields for core functionality of object-roles"

    class Meta:
        abstract = True

    # role_definition set on child models to set appropriate help_text and related_name
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, help_text=_("The content type of the subject of permission assignments. Duplicated from related RoleDefinition.")
    )
    object_id = models.TextField(null=False, help_text=_("The database primary key of the subject of permission assignments."))
    content_object = GenericForeignKey('content_type', 'object_id')

    @classmethod
    def _visible_items(cls, eval_cls, user, qs=None):
        permission_qs = eval_cls.objects.filter(
            role__in=user.has_roles.all(),
            content_type_id=models.OuterRef('content_type_id'),
        )
        # NOTE: type casting is necessary in postgres but not sqlite3
        object_id_field = cls._meta.get_field('object_id')
        obj_filter = models.Q(object_id__in=permission_qs.values_list(Cast('object_id', output_field=object_id_field)))

        if not hasattr(user, '_singleton_permission_objs'):
            user._singleton_permission_objs = RoleDefinition.user_global_permissions(user)

        if qs is None:
            qs = cls.objects.all()

        if user._singleton_permission_objs:
            super_ct_ids = set(perm.content_type_id for perm in user._singleton_permission_objs)
            # content_type=None condition: A good-enough rule - you can see other global assignments if you have any yourself
            return qs.filter(obj_filter | models.Q(content_type__in=super_ct_ids) | models.Q(content_type=None))
        return qs.filter(obj_filter)

    @classmethod
    def visible_items(cls, user, qs=None):
        "This ORs querysets to show assignments to both UUID and integer pk models"
        return cls._visible_items(RoleEvaluation, user, qs) | cls._visible_items(RoleEvaluationUUID, user, qs)

    @property
    def cache_id(self):
        "The ObjectRole GenericForeignKey is text, but cache needs to match models"
        return RoleEvaluation._meta.get_field('object_id').to_python(self.object_id)


class AssignmentBase(ImmutableCommonModel, ObjectRoleFields):
    """
    This uses some parts of CommonModel to save metadata like documenting
    the user who assigned the permission and timestamp when it happened.
    This caches ObjectRole fields for purposes of serializers,
    both models are immutable, making caching easy.
    """

    object_role = models.ForeignKey(
        'dab_rbac.ObjectRole', on_delete=models.CASCADE, editable=False, null=True, help_text=_("A roll-up of the fields (role_definition, content_type).")
    )
    object_id = models.TextField(
        null=True, blank=True, help_text=_('The primary key of the object this assignment applies to; null value indicates system-wide assignment.')
    )
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True, help_text=_("The content type this applies to."))

    # object_role is internal, and not shown in serializer
    # content_type does not have a link, and ResourceType will be used in lieu sometime
    ignore_relations = ['content_type', 'object_role']

    class Meta:
        app_label = 'dab_rbac'
        abstract = True

    def save(self, *args, **kwargs):
        if not self.id:  # usually only new objects can be saved, but super needs to do error handling
            if self.object_role_id and (not self.object_id):
                # Cache fields from the associated object_role
                self.object_id = self.object_role.object_id
                self.content_type_id = self.object_role.content_type_id
                self.role_definition_id = self.object_role.role_definition_id
        return super().save(*args, **kwargs)


class RoleUserAssignment(AssignmentBase):
    role_definition = models.ForeignKey(
        RoleDefinition,
        on_delete=models.CASCADE,
        help_text=_("The role definition which defines permissions conveyed by this assignment."),
        related_name='user_assignments',
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='role_assignments', help_text=_("The user this role is assigned to.")
    )
    router_basename = 'roleuserassignment'

    class Meta:
        app_label = 'dab_rbac'
        ordering = ['id']
        unique_together = ('user', 'object_role')

    def __repr__(self):
        return f'RoleUserAssignment(pk={self.id})'

    @property
    def actor(self):
        "Really simple helper to give same behavior between user and role assignments"
        return self.user


class RoleTeamAssignment(AssignmentBase):
    role_definition = models.ForeignKey(
        RoleDefinition,
        on_delete=models.CASCADE,
        help_text=_("The role definition which defines permissions conveyed by this assignment."),
        related_name='team_assignments',
    )
    team = models.ForeignKey(
        settings.ANSIBLE_BASE_TEAM_MODEL, on_delete=models.CASCADE, related_name='role_assignments', help_text=_("The team that receives permissions.")
    )
    router_basename = 'roleteamassignment'

    class Meta:
        app_label = 'dab_rbac'
        ordering = ['id']
        unique_together = ('team', 'object_role')

    def __repr__(self):
        return f'RoleTeamAssignment(pk={self.id})'

    @property
    def actor(self):
        "Really simple helper to give same behavior between user and role assignments"
        return self.team


class ObjectRole(ObjectRoleFields):
    """
    This is the successor to the Role model in the old AWX RBAC system
    It is renamed to ObjectRole to distinguish from the abstract or generic
    RoleDefinition which does not apply to a particular object.

    This matches the RoleDefinition to a content_object.
    After this is created, users and teams can be added to gives those
    permissions to that user or team, for that content_object
    """

    class Meta:
        app_label = 'dab_rbac'
        verbose_name_plural = _('object_roles')
        indexes = [models.Index(fields=["content_type", "object_id"])]
        ordering = ("content_type", "object_id")
        constraints = [models.UniqueConstraint(name='one_object_role_per_object_and_role', fields=['object_id', 'content_type', 'role_definition'])]

    role_definition = models.ForeignKey(
        RoleDefinition,
        on_delete=models.CASCADE,
        help_text=_("The role definition which defines what permissions this object role grants."),
        related_name='object_roles',
    )
    users = models.ManyToManyField(
        to=settings.AUTH_USER_MODEL,
        through='dab_rbac.RoleUserAssignment',
        through_fields=("object_role", "user"),
        related_name='has_roles',
        help_text=_("Users who have access to the permissions defined by this object role."),
    )
    teams = models.ManyToManyField(
        to=settings.ANSIBLE_BASE_TEAM_MODEL,
        through='dab_rbac.RoleTeamAssignment',
        through_fields=("object_role", "team"),
        related_name='has_roles',
        help_text=_("Teams or groups who have access to the permissions defined by this object role."),
    )
    # COMPUTED DATA
    provides_teams = models.ManyToManyField(
        settings.ANSIBLE_BASE_TEAM_MODEL,
        related_name='member_roles',
        editable=False,
        help_text=_("Users who have this role obtain member access to these teams, and inherit all their permissions."),
    )

    def __str__(self):
        return f'ObjectRole(pk={self.id}, {self.content_type.model}={self.object_id})'

    def save(self, *args, **kwargs):
        if self.id:
            raise RuntimeError('ObjectRole model is immutable, use RoleDefinition.give_permission method')
        return super().save(*args, **kwargs)

    def summary_fields(self):
        return {'id': self.id}

    def descendent_roles(self):
        "Returns a set of roles that you implicitly have if you have this role"
        descendents = set()
        for target_team in self.provides_teams.all():
            # the roles that offer these permissions could change as a result of adding teams
            descendents.update(set(target_team.has_roles.all()))
        return descendents

    def expected_direct_permissions(self, types_prefetch=None):
        expected_evaluations = set()
        cached_id_lists = {}
        if not types_prefetch:
            types_prefetch = TypesPrefetch()
        role_content_type = types_prefetch.get_content_type(self.content_type_id)
        role_model = role_content_type.model_class()
        # ObjectRole.object_id is stored as text, we convert it to the model pk native type
        object_id = role_model._meta.pk.to_python(self.object_id)
        for permission in types_prefetch.permissions_for_object_role(self):
            permission_content_type = types_prefetch.get_content_type(permission.content_type_id)

            # direct object permission
            if permission.content_type_id == self.content_type_id:
                expected_evaluations.add((permission.codename, self.content_type_id, object_id))
                continue

            # add child permission on the parent object, usually only for add permission
            if is_add_perm(permission.codename) or settings.ANSIBLE_BASE_CACHE_PARENT_PERMISSIONS:
                expected_evaluations.add((permission.codename, self.content_type_id, object_id))

            # add child object permission on child objects
            # Only propogate add permission to children which are parents of the permission model
            filter_path = None
            child_model = None
            if is_add_perm(permission.codename):
                for path, model in permission_registry.get_child_models(role_model):
                    if '__' in path and model._meta.model_name == permission_content_type.model:
                        path_to_parent, filter_path = path.split('__', 1)
                        child_model = permission_content_type.model_class()._meta.get_field(path_to_parent).related_model
                        eval_ct = ContentType.objects.get_for_model(child_model).id
                if not child_model:
                    continue
            else:
                for path, model in permission_registry.get_child_models(role_model):
                    if model._meta.model_name == permission_content_type.model:
                        filter_path = path
                        child_model = model
                        eval_ct = permission.content_type_id
                        break
                else:
                    logger.warning(f'{self.role_definition} listed {permission.codename} but model is not a child, ignoring')
                    continue

            # fetching child objects of an organization is very performance sensitive
            # for multiple permissions of same type, make sure to only do query once
            id_list = []
            if eval_ct in cached_id_lists:
                id_list = cached_id_lists[eval_ct]
            else:
                id_list = child_model.objects.filter(**{filter_path: object_id}).values_list('pk', flat=True)
                cached_id_lists[eval_ct] = list(id_list)

            for id in id_list:
                expected_evaluations.add((permission.codename, eval_ct, id))
        return expected_evaluations

    def needed_cache_updates(self, types_prefetch=None):
        existing_partials = dict()
        for permission_partial in self.permission_partials.all():
            existing_partials[permission_partial.obj_perm_id()] = permission_partial
        for permission_partial in self.permission_partials_uuid.all():
            existing_partials[permission_partial.obj_perm_id()] = permission_partial

        expected_evaluations = self.expected_direct_permissions(types_prefetch)

        for team in self.provides_teams.all():
            for team_role in team.has_roles.all():
                expected_evaluations.update(team_role.expected_direct_permissions(types_prefetch))

        existing_set = set(existing_partials.keys())

        to_delete = set()
        for identifier in existing_set - expected_evaluations:
            to_delete.add((existing_partials[identifier].id, type(identifier[-1])))

        to_add = []
        for codename, ct_id, obj_pk in expected_evaluations - existing_set:
            to_add.append(RoleEvaluation(codename=codename, content_type_id=ct_id, object_id=obj_pk, role=self))

        return (to_delete, to_add)


class RoleEvaluationMeta:
    app_label = 'dab_rbac'
    verbose_name_plural = _('role_object_permissions')
    indexes = [
        models.Index(fields=["role", "content_type_id", "object_id"]),  # used by get_roles_on_resource
        models.Index(fields=["role", "content_type_id", "codename"]),  # used by accessible_objects
    ]
    constraints = [models.UniqueConstraint(name='one_entry_per_object_permission_and_role', fields=['object_id', 'content_type_id', 'codename', 'role'])]


# COMPUTED DATA
class RoleEvaluationFields(models.Model):
    """
    Cached data that shows what permissions an ObjectRole gives its owners
    example:
        ObjectRole 423 gives users execute access to job template 37

    RoleAncestorEntry model in old AWX RBAC system is a direct analog

    This is used to make permission evaluations via querysets returning object ids
    the data in this table is created from the ObjectRole and RoleDefinition data
      you should not interact with this table yourself
      the only method that should ever write to this table is
        compute_object_role_permissions()

    In the above example, "ObjectRole 423" may be a role that grants membership
    to a team, and that team was given permission to another ObjectRole.
    """

    class Meta:
        abstract = True

    def __str__(self):
        return (
            f'{self._meta.verbose_name.title()}(pk={self.id}, codename={self.codename}, object_id={self.object_id}, '
            f'content_type_id={self.content_type_id}, role_id={self.role_id})'
        )

    def save(self, *args, **kwargs):
        if self.id:
            raise RuntimeError(f'{self._meta.model_name} model is immutable and only used internally')
        return super().save(*args, **kwargs)

    codename = models.TextField(null=False, help_text=_("The name of the permission, giving the action and the model, from the Django Permission model."))
    # NOTE: we do not form object_id and content_type into a content_object, following from AWX practice
    # this can be relaxed as we have comparative performance testing to confirm doing so does not affect permissions
    content_type_id = models.PositiveIntegerField(null=False, help_text=_("The related content type id."))

    def obj_perm_id(self):
        "Used for in-memory hashing of the type of object permission this represents"
        return (self.codename, self.content_type_id, self.object_id)

    @classmethod
    def accessible_ids(cls, model_cls, actor, codename: str, content_types: Optional[Iterable[int]] = None, cast_field=None) -> QuerySet:
        """
        Corresponds to AWX accessible_pk_qs

        Use instead of `MyModel.objects` when you want to only consider
        resources that a user has specific permissions for. For example:
        MyModel.accessible_objects(user, 'view_mymodel').filter(name__istartswith='bar')

        Intended to be used for users, but should also be valid for teams
        """
        # We only have a content_types exception for multiple content types for polymorphic models
        # for normal models you should not need it, but AWX unified_ models need it to get by
        filter_kwargs = dict(role__in=actor.has_roles.all(), codename=codename)
        if content_types:
            filter_kwargs['content_type_id__in'] = content_types
        else:
            filter_kwargs['content_type_id'] = ContentType.objects.get_for_model(model_cls).id
        qs = cls.objects.filter(**filter_kwargs)
        if cast_field is None:
            return qs.values_list('object_id').distinct()
        else:
            return qs.values_list(Cast('object_id', output_field=cast_field)).distinct()

    @classmethod
    def accessible_objects(cls, model_cls, user, codename, queryset: Optional[QuerySet] = None) -> QuerySet:
        if queryset is None:
            queryset = model_cls.objects.all()
        return queryset.filter(pk__in=cls.accessible_ids(model_cls, user, codename))

    @classmethod
    def get_permissions(cls, user, obj):
        """
        Returns permissions that a user has to obj from object-roles,
        does not consider permissions from user flags or system-wide roles
        """
        return cls.objects.filter(role__in=user.has_roles.all(), content_type_id=ContentType.objects.get_for_model(obj).id, object_id=obj.id).values_list(
            'codename', flat=True
        )

    @classmethod
    def has_obj_perm(cls, user, obj, codename) -> bool:
        """
        Note this behaves similar in function to the REST Framework has_object_permission
        method on permission classes, but it is named differently to avoid unintentionally conflicting
        """
        return cls.objects.filter(
            role__in=user.has_roles.all(), content_type_id=ContentType.objects.get_for_model(obj).id, object_id=obj.pk, codename=codename
        ).exists()


class RoleEvaluation(RoleEvaluationFields):
    class Meta(RoleEvaluationMeta):
        pass

    role = models.ForeignKey(
        ObjectRole,
        null=False,
        on_delete=models.CASCADE,
        related_name='permission_partials',
        help_text=_("The object role that grants this form of permission."),
    )
    object_id = models.PositiveIntegerField(null=False, help_text=_("The id of the object that the related role gives the related permission to."))


class RoleEvaluationUUID(RoleEvaluationFields):
    "Cache for UUID type models"

    class Meta(RoleEvaluationMeta):
        constraints = [
            models.UniqueConstraint(name='one_entry_per_object_permission_and_role_uuid', fields=['object_id', 'content_type_id', 'codename', 'role'])
        ]

    role = models.ForeignKey(
        ObjectRole,
        null=False,
        on_delete=models.CASCADE,
        related_name='permission_partials_uuid',
        help_text=_("The object role that grants this form of permission."),
    )
    object_id = models.UUIDField(null=False, help_text=_("The object UUID this role evaluation will be applied to."))


def get_evaluation_model(cls):
    pk_field = cls._meta.pk
    # For proxy models, including django-polymorphic, use the id field from parent table
    # we accomplish this by inspecting the raw database type of the field
    pk_db_type = pk_field.db_type(connection)
    for eval_cls in (RoleEvaluation, RoleEvaluationUUID):
        if pk_db_type == eval_cls._meta.get_field('object_id').db_type(connection):
            return eval_cls
    # HACK: integer pk caching is handled by same model for now, better to use default pk type later
    # the integer unsigned case happens in AWX in sqlite3 specifically
    if pk_db_type in ('bigint', 'integer', 'integer unsigned'):
        return RoleEvaluation

    raise RuntimeError(f'Model {cls._meta.model_name} primary key type of {type(pk_field)} (db type {pk_db_type}) is not supported')



from rest_framework import authentication


class SessionAuthentication(authentication.SessionAuthentication):
    """
    This class allows us to fail with a 401 if the user is not authenticated.
    """

    def authenticate_header(self, request):
        return "Session"



import logging

from django.contrib.contenttypes.models import ContentType
from django.db.models.query import IntegrityError

from ansible_base.jwt_consumer.common.auth import JWTAuthentication
from ansible_base.jwt_consumer.common.exceptions import InvalidService
from ansible_base.rbac.models import RoleDefinition, RoleUserAssignment
from ansible_base.resource_registry.models import Resource

logger = logging.getLogger('ansible_base.jwt_consumer.hub.auth')


class HubJWTAuth(JWTAuthentication):

    def get_galaxy_models(self):
        '''This is separate from process_permissions purely for testability.'''
        try:
            from galaxy_ng.app.models import Organization, Team
        except ImportError:
            raise InvalidService("automation-hub")

        return Organization, Team

    def process_permissions(self):
        # Map teams in the JWT to Automation Hub groups.
        Organization, Team = self.get_galaxy_models()
        self.team_content_type = ContentType.objects.get_for_model(Team)
        self.org_content_type = ContentType.objects.get_for_model(Organization)

        # TODO - galaxy does not have an org admin roledef yet
        # admin_orgs = []

        # TODO - galaxy does not have an org member roledef yet
        # member_orgs = []

        # The "shared" [!local] teams this user admins
        admin_teams = []

        # the teams this user should have a "shared" [!local] assignment to
        member_teams = []

        for role_name in self.common_auth.token.get('object_roles', {}).keys():
            if role_name.startswith('Team'):
                for object_index in self.common_auth.token['object_roles'][role_name]['objects']:
                    team_data = self.common_auth.token['objects']['team'][object_index]
                    ansible_id = team_data['ansible_id']
                    try:
                        team = Resource.objects.get(ansible_id=ansible_id).content_object
                    except Resource.DoesNotExist:
                        try:
                            team = self.common_auth.get_or_create_resource('team', team_data)[1]
                        except IntegrityError as e:
                            logger.warning(
                                f"Got integrity error ({e}) on {team_data}. Skipping team assignment. "
                                "Please make sure the sync task is running to prevent this warning in the future."
                            )
                            continue

                    if role_name == 'Team Admin':
                        admin_teams.append(team)
                    elif role_name == 'Team Member':
                        member_teams.append(team)

        for roledef_name, teams in [('Team Admin', admin_teams), ('Team Member', member_teams)]:

            # the "shared" "non-local" definition ...
            roledef = RoleDefinition.objects.get(name=roledef_name)

            # pks for filtering ...
            team_pks = [team.pk for team in teams]

            # delete all assignments not defined by this jwt ...
            for assignment in RoleUserAssignment.objects.filter(user=self.common_auth.user, role_definition=roledef).exclude(object_id__in=team_pks):
                team = Team.objects.get(pk=assignment.object_id)
                roledef.remove_permission(self.common_auth.user, team)

            # assign "non-local" for each team ...
            for team in teams:
                roledef.give_permission(self.common_auth.user, team)

        auditor_roledef = RoleDefinition.objects.get(name='Platform Auditor')
        if "Platform Auditor" in self.common_auth.token.get('global_roles', []):
            auditor_roledef.give_global_permission(self.common_auth.user)
        else:
            auditor_roledef.remove_global_permission(self.common_auth.user)



import logging

from django.apps import apps
from django.conf import settings
from django.db.models import Model
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import SAFE_METHODS, BasePermission, DjangoObjectPermissions

from ansible_base.lib.utils.models import is_add_perm
from ansible_base.lib.utils.settings import get_setting
from ansible_base.rbac import permission_registry
from ansible_base.rbac.evaluations import has_super_permission
from ansible_base.rbac.policies import can_change_user

logger = logging.getLogger('ansible_base.rbac.api.permissions')


class IsSystemAdminOrAuditor(BasePermission):
    """
    Allows write access only to system admin users.
    Allows read access only to system auditor users.
    """

    def has_permission(self, request, view):
        if not (request.user and request.user.is_authenticated):
            return False
        if request.method in SAFE_METHODS:
            return has_super_permission(request.user, 'view')
        return has_super_permission(request.user)


class AuthenticatedReadAdminChange(IsSystemAdminOrAuditor):
    "Any authenticated user can view, but only admin users can do CRUD"

    def has_permission(self, request, view):
        if not (request.user and request.user.is_authenticated):
            return False
        if request.method in SAFE_METHODS:
            return True
        return has_super_permission(request.user)


def is_cloned_request(request) -> bool:
    """Tells whether this is a fake request

    The DRF API browser and schema generators call permission methods
    multiple times for form generation purposes.
    In these cases, a request wrapps the original and the method will not match.
    """
    return bool(request.method != request._request.method)


class RoleDefinitionPermissions(AuthenticatedReadAdminChange):
    def has_permission(self, request, view):
        if is_cloned_request(request) and settings.ANSIBLE_BASE_ALLOW_CUSTOM_ROLES is False:
            return False
        return super().has_permission(request, view)


class AnsibleBaseObjectPermissions(DjangoObjectPermissions):
    def has_create_permission(self, request, model_cls) -> bool:
        "Does the request user absolutely have permission to create an object via superuser or system role"
        full_codename = f'add_{model_cls._meta.model_name}'
        return has_super_permission(request.user, full_codename)

    def abstract_create_permission(self, request, model_cls) -> bool:
        "Abstractly, without knowing the request data, could the request user theoretically create an object"
        parent_model = permission_registry.get_parent_model(model_cls)
        if parent_model is None:
            return self.has_create_permission(request, model_cls)
        else:
            if request.user.is_superuser:
                return True  # special case only for superuser when no objects exist
            return parent_model.access_qs(request.user, f'add_{model_cls._meta.model_name}').exists()

    def has_permission(self, request, view):
        "Some of this comes from ModelAccessPermission. We assume user.permissions is unused"
        if not request.user or (not request.user.is_authenticated and self.authenticated_users_only):
            return False

        # Workaround to ensure DjangoModelPermissions are not applied
        # to the root view when using DefaultRouter.
        if getattr(view, '_ignore_model_permissions', False):
            return True

        # Following is DAB RBAC specific, handle add permission checking
        if request.method == 'POST' and view.action == 'create':
            model_cls = self._queryset(view).model
            parent_field_name = permission_registry.get_parent_fd_name(model_cls)
            if parent_field_name is None:
                result = self.has_create_permission(request, model_cls)
                if (not result) and (not is_cloned_request(request)):
                    logger.warning(f'User {request.user.pk} lacks global add_{model_cls._meta.model_name} permission to create {model_cls._meta.model_name}')
                return result
        elif request.method == 'POST' and is_cloned_request(request):
            # If this is OPTIONS purposes
            # return a speculative answer about whether user might be generally able to create
            model_cls = self._queryset(view).model
            return self.abstract_create_permission(request, model_cls)

        # We are not checking many things here, a GET to list views can return 0 objects
        return True

    def get_required_object_permissions(self, method, model_cls, view=None):
        special_action = getattr(view, 'rbac_action', None)
        if special_action:
            return [f'{special_action}_{model_cls._meta.model_name}']
        perms = super().get_required_object_permissions(method, model_cls)
        # Remove add permissions, since they are handled in has_permission
        return [p for p in perms if not is_add_perm(p)]

    def has_object_permission_by_codename(self, request, obj, perms):
        return all(request.user.has_obj_perm(obj, perm) for perm in perms)

    def model_is_valid(self, model_cls):
        return permission_registry.is_registered(model_cls)

    def has_object_permission(self, request, view, obj):
        if not isinstance(obj, Model):
            logger.info(f'Object-permission check called with non-object {type(obj)} for {request.method}')
            return True  # for the DRF browsable API, showing PATCH form field

        queryset = self._queryset(view)
        model_cls = queryset.model
        if not self.model_is_valid(model_cls):
            if request.user.is_superuser:
                return True
            logger.warning(f'User {request.user.pk} denied {request.method} to {obj._meta.model_name}, not in DAB RBAC permission registry')
            raise Http404

        perms = self.get_required_object_permissions(request.method, model_cls, view=view)

        if not self.has_object_permission_by_codename(request, obj, perms):
            # If the user does not have permissions we need to determine if
            # they have read permissions to see 403, or not, and simply see
            # a 404 response.
            if not is_cloned_request(request):
                logger.warning(f'User {request.user.pk} lacks {perms} permission to obj {obj._meta.model_name}-{obj.pk}')

            if request.method in SAFE_METHODS:
                # Read permissions already checked and failed, no need
                # to make another lookup.
                raise Http404

            read_perms = self.get_required_object_permissions('GET', model_cls)
            if not all(request.user.has_obj_perm(obj, perm) for perm in read_perms):
                raise Http404

            # Has read permissions.
            return False

        return True


class AnsibleBaseUserPermissions(AnsibleBaseObjectPermissions):
    def model_is_valid(self, model_cls):
        return bool(model_cls._meta.model_name == 'user')

    def has_create_permission(self, request, model_cls):
        org_cls = apps.get_model(settings.ANSIBLE_BASE_ORGANIZATION_MODEL)

        if request.user.is_superuser:
            return True

        if not get_setting('MANAGE_ORGANIZATION_AUTH', False):
            return False

        return org_cls.access_qs(request.user, 'change_organization').exists()

    def abstract_create_permission(self, request, model_cls):
        # For users, only, ability to create a new user does not depend on the data
        return self.has_create_permission(request, model_cls)

    def has_object_permission_by_codename(self, request, obj, perms):
        if perms:
            if request.method == 'DELETE' and request.user.pk == obj.pk:
                raise PermissionDenied({'detail': _("You can't delete yourself")})
            return can_change_user(request.user, obj)
        return True



# Generated by Django 4.2.8 on 2024-04-09 17:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0006_alter_authenticatoruser_unique_together'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authenticator',
            name='users',
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='provider',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='authenticator_provider', to='dab_authentication.authenticator', to_field='slug'),
        ),
    ]



from rest_framework.response import Response

from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_class, get_authenticator_plugins
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView


class AuthenticatorPluginView(AnsibleBaseDjangoAppApiView):
    def get(self, request, format=None):
        plugins = get_authenticator_plugins()
        resp = {"authenticators": []}

        for p in plugins:
            try:
                klass = get_authenticator_class(p)
                config = klass.configuration_class()
                config_schema = config.get_configuration_schema()
                resp['authenticators'].append(
                    {"type": p, "configuration_schema": config_schema, "documentation_url": getattr(config, "documentation_url", None)}
                )
            except ImportError as ie:
                # If we got an import error its already logged and we can move on
                if 'errors' not in resp:
                    resp['errors'] = []
                resp['errors'].append(ie.__str__())

        resp['authenticators'] = sorted(resp['authenticators'], key=lambda k: k['type'])

        return Response(resp)



# Generated by Django 4.2.11 on 2024-08-08 19:52

import ansible_base.resource_registry.models.service_identifier
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_resource_registry', '0004_remove_resourcetype_migrated'),
    ]

    operations = [
        migrations.AddField(
            model_name='resource',
            name='is_partially_migrated',
            field=models.BooleanField(default=False, help_text="This gets set to True when a resource has been copied into the resource server, but the service_id hasn't been updated yet."),
        ),
        migrations.AlterField(
            model_name='resource',
            name='service_id',
            field=models.UUIDField(default=ansible_base.resource_registry.models.service_identifier.service_id, help_text='ID of the service responsible for managing this resource.'),
        ),
    ]



from django.apps import AppConfig


class FeatureFlagsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.feature_flags'
    label = 'dab_feature_flags'
    verbose_name = 'Feature Flags'



# Generated by Django 4.2.6 on 2024-03-02 14:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_resource_registry', '0002_remove_resource_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='resource',
            name='object_id',
            field=models.TextField(db_index=True),
        ),
    ]



"""
Utilities for Python collections
"""


def first_matching(predicate, collection, default=StopIteration("No item matched the predicate.")):
    """
    Return the first element from `collection` that satisfies `predicate`.
    If `default` is an exception, raise it if no element satisfies `predicate`.
    If `default` is not an exception, return it if no element satisfies `predicate`.

    :param predicate: A function that takes an element from `collection` and returns a boolean.
    :param collection: An iterable.
    :param default: The value to return if no element satisfies `predicate`.
    :return: The first element from `collection` that satisfies `predicate`.
             If no element satisfies `predicate`, return `default`, or raise an exception if `default` is an exception.
    """

    for i in collection:
        if predicate(i):
            return i

    if isinstance(default, Exception):
        raise default

    return default



# Generated by Django 4.2.11 on 2024-04-19 12:39

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_authentication', '0008_remove_authenticator_users_unique'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticatoruser',
            name='provider',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='authenticator_providers', to='dab_authentication.authenticator', to_field='slug'),
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authenticator_users', to=settings.AUTH_USER_MODEL),
        ),
    ]



import logging
from functools import lru_cache
from glob import glob
from os.path import basename, isfile, join
from typing import Optional

from django.conf import settings
from django.db.models.fields import uuid
from django.utils.text import slugify

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.utils')
setting = 'ANSIBLE_BASE_AUTHENTICATOR_CLASS_PREFIXES'


@lru_cache
def get_authenticator_plugins() -> list:
    class_prefixes = getattr(settings, setting, [])
    plugins = []
    for class_prefix in class_prefixes:
        path_info = class_prefix.split('.')
        last_class_path = path_info[-1]
        parent_class = __import__(class_prefix, globals(), locals(), [last_class_path], 0)
        for path in parent_class.__path__:
            for file in glob(join(path, "*.py")):
                file_name = basename(file)
                if isfile(file) and file_name not in ['__init__.py', 'utils.py', 'base.py'] and not file_name.startswith('_'):
                    plugins.append(f"{class_prefix}.{file_name.replace('.py', '')}")
    return plugins


def get_authenticator_class(authenticator_type: str):
    if not authenticator_type:
        raise ImportError("Must pass authenticator type to import")
    try:
        logger.debug(f"Attempting to load class {authenticator_type}")
        auth_class = __import__(authenticator_type, globals(), locals(), ['AuthenticatorPlugin'], 0)
        return auth_class.AuthenticatorPlugin
    except Exception as e:
        logger.exception(f"The specified authenticator type {authenticator_type} could not be loaded, see exception below")
        raise ImportError(f"The specified authenticator type {authenticator_type} could not be loaded") from e


def get_authenticator_plugin(authenticator_type: str):
    AuthClass = get_authenticator_class(authenticator_type)
    return AuthClass()


def get_authenticator_urls(authenticator_type: str) -> list:
    try:
        urls_module = __import__(authenticator_type, globals(), locals(), ['urls'], 0)
        return getattr(urls_module, 'urls', [])
    except Exception as e:
        logger.error(f"Failed to load urls from {authenticator_type} {e}")
    return []


def generate_authenticator_slug(value: Optional[str] = None) -> str:
    if not value:
        value = str(uuid.uuid4())
    return slugify(value)



# Generated by Django 4.2.16 on 2024-11-22 21:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('dab_activitystream', '0004_alter_entry_created_alter_entry_created_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='entry',
            name='changes',
            field=models.JSONField(blank=True, help_text='The changes to the system recorded by this entry.', null=True),
        ),
        migrations.AlterField(
            model_name='entry',
            name='content_type',
            field=models.ForeignKey(help_text='The content type which was changed in this entry.', on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='entry',
            name='object_id',
            field=models.TextField(blank=True, help_text='The object id which was modified in this entry.', null=True),
        ),
        migrations.AlterField(
            model_name='entry',
            name='operation',
            field=models.CharField(choices=[('create', 'Entity created'), ('update', 'Entity updated'), ('delete', 'Entity deleted'), ('associate', 'Entity was associated with another entity'), ('disassociate', 'Entity was disassociated with another entity')], help_text='The type of change recorded by the entry (i.e. create/update/delete/etc).', max_length=12),
        ),
        migrations.AlterField(
            model_name='entry',
            name='related_content_type',
            field=models.ForeignKey(blank=True, help_text='The content type if this entry is recording an many to many (M2M) change.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='related_content_type', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='entry',
            name='related_field_name',
            field=models.CharField(blank=True, help_text='The related field name if this entry is for an many to many (M2M) relationship.', max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='entry',
            name='related_object_id',
            field=models.TextField(blank=True, help_text='The object id of the related model if this entry is for an many to many (M2M) change.', null=True),
        ),
    ]



import logging
import time
from base64 import b64encode

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding

from ansible_base.jwt_consumer.common.cert import JWTCert, JWTCertException
from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.jwt_consumer.common.util')

_SHARED_SECRET = 'trusted_proxy'


def generate_x_trusted_proxy_header(key: str) -> str:
    private_key = serialization.load_pem_private_key(bytes(key, 'utf-8'), password=None)
    timestamp = time.time_ns()
    message = f'{_SHARED_SECRET}-{timestamp}'
    signature = private_key.sign(bytes(message, 'utf-8'), padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH), hashes.SHA256())
    return f"{timestamp}-{signature.hex()}"


def validate_x_trusted_proxy_header(header_value: str, ignore_cache=False) -> bool:
    try:
        cert = JWTCert()
        cert.get_decryption_key(ignore_cache=ignore_cache)
        if cert.key is None:
            raise JWTCertException(f"Failed to load cert from setting {cert.key_name}")
    except JWTCertException as e:
        logger.error(f"Failed to validate x-trusted-proxy-header, unable to load cert {e}")
        return False

    try:
        public_key = serialization.load_pem_public_key(cert.key.encode('utf-8'))
    except Exception:
        logger.exception("Failed to load public key")
        return False

    try:
        timestamp, signature = header_value.split('-', maxsplit=1)
    except ValueError:
        logger.warning("Failed to validate x-trusted-proxy-header, malformed, expected value to contain a -")
        return False

    # Validate that the header has been cut within the last 300ms (by default)
    try:
        if time.time_ns() - int(timestamp) > get_setting('trusted_header_timeout_in_ns', 300000000):
            logger.warning(f"Timestamp {timestamp} was too old to be valid alter trusted_header_timeout_in_ns if needed")
            return False
    except ValueError:
        logger.warning(f"Unable to convert timestamp (base64) {b64encode(timestamp.encode('UTF-8'))} into an integer")
        return False

    try:
        signature_bytes = bytes.fromhex(signature)
    except ValueError:
        logger.warning("Failed to validate x-trusted-proxy-header, malformed, expected signature to well-formed base64")
        return False

    try:
        public_key.verify(
            signature_bytes,
            bytes(f'{_SHARED_SECRET}-{timestamp}', 'utf-8'),
            padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
            hashes.SHA256(),
        )
        return True
    except InvalidSignature:
        if ignore_cache or not cert.cached:
            return False
        return validate_x_trusted_proxy_header(header_value, ignore_cache=True)



import logging

from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.utils.authentication import determine_username_from_uid, get_or_create_authenticator_user
from ansible_base.authentication.utils.claims import update_user_claims

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.local')

# TODO: Change the validator to not allow it to be deleted or a second one added


class LocalConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://docs.djangoproject.com/en/4.2/ref/contrib/auth/#django.contrib.auth.backends.ModelBackend"

    def validate(self, data):
        if data != {}:
            raise ValidationError(_({"configuration": "Can only be {} for local authenticators"}))
        return data


class AuthenticatorPlugin(ModelBackend, AbstractAuthenticatorPlugin):
    configuration_class = LocalConfiguration
    logger = logger
    type = "local"
    category = "password"

    def __init__(self, database_instance=None, *args, **kwargs):
        super().__init__(database_instance, *args, **kwargs)

    def authenticate(self, request, username=None, password=None, **kwargs):
        if not username or not password:
            return None

        if not self.database_instance:
            return None

        if not self.database_instance.enabled:
            logger.info(f"Local authenticator {self.database_instance.name} is disabled, skipping")
            return None

        # Determine the user name for this authenticator, we have to call this so that we can "attach" to a pre-created user
        new_username = determine_username_from_uid(username, self.database_instance)
        # However we can't really accept a different username because we are the local authenticator imageine if:
        #    User "a" is from another authenticator and has an AuthenticatorUser
        #    User "a" tried to login from local authenticator
        #    The above function will return a username of "a<hash>"
        #    We then try to do local authentication with the database from a different username that will not exist in the database, so it would never work
        if new_username != username:
            return None

        user = super().authenticate(request, username, password, **kwargs)

        # This auth class doesn't create any new local users, but we still need to make sure
        # it has an AuthenticatorUser associated with it.
        if user:
            get_or_create_authenticator_user(
                username,
                authenticator=self.database_instance,
                user_details={},
                extra_data={
                    "username": username,
                    "first_name": user.first_name,
                    "last_name": user.last_name,
                    "email": user.email,
                    "is_superuser": user.is_superuser,
                },
            )

        return update_user_claims(user, self.database_instance, [])



import re
from collections import defaultdict
from typing import Optional, Type, Union

from django.conf import settings
from django.db.models import Model
from rest_framework.exceptions import ValidationError

from ansible_base.lib.utils.models import is_add_perm
from ansible_base.rbac.permission_registry import permission_registry


def system_roles_enabled():
    return bool(
        settings.ANSIBLE_BASE_ALLOW_SINGLETON_ROLES_API
        and (settings.ANSIBLE_BASE_ALLOW_SINGLETON_USER_ROLES or settings.ANSIBLE_BASE_ALLOW_SINGLETON_TEAM_ROLES)
    )


def prnt_model_name(model: Optional[Type[Model]]) -> str:
    return model._meta.model_name if model else 'global role'


def prnt_codenames(codename_set: set[str]) -> str:
    return ', '.join(codename_set)


def codenames_for_cls(cls) -> list[str]:
    "Helper method that gives the Django permission codenames for a given class"
    return [t[0] for t in cls._meta.permissions] + [f'{act}_{cls._meta.model_name}' for act in cls._meta.default_permissions]


def permissions_allowed_for_system_role() -> dict[Type[Model], list[str]]:
    "Permission codenames useable in system-wide roles, which have content_type set to None"
    permissions_by_model = defaultdict(list)
    for cls in sorted(permission_registry.all_registered_models, key=lambda cls: cls._meta.model_name):
        is_team = bool(cls._meta.model_name == 'team')
        for codename in codenames_for_cls(cls):
            if is_team and (not codename.startswith('view')):
                continue  # special exclusion of team object permissions from system-wide roles
            permissions_by_model[cls].append(codename)
    return permissions_by_model


def permissions_allowed_for_role(cls) -> dict[Type[Model], list[str]]:
    "Permission codenames valid for a RoleDefinition of given class, organized by permission class"
    if cls is None:
        return permissions_allowed_for_system_role()

    if not permission_registry.is_registered(cls):
        raise ValidationError(f'Django-ansible-base RBAC does not track permissions for model {cls._meta.model_name}')

    # Include direct model permissions (except for add permission)
    permissions_by_model = defaultdict(list)
    permissions_by_model[cls] = [codename for codename in codenames_for_cls(cls) if not is_add_perm(codename)]

    # Include model permissions for all child models, including the add permission
    for rel, child_cls in permission_registry.get_child_models(cls):
        permissions_by_model[child_cls] += codenames_for_cls(child_cls)

    return permissions_by_model


def combine_values(data: dict[Type[Model], list[str]]) -> set[str]:
    "Utility method to merge everything in .values() into a single set"
    ret = set()
    for this_list in data.values():
        ret |= set(this_list)
    return ret


def validate_role_definition_enabled(permissions, content_type) -> None:
    """Called by API and managers, raises exception if settings allow this role type

    Like the similar method for assignments, this policies the ANSIBLE_BASE_ALLOW_ settings
    """
    if not settings.ANSIBLE_BASE_ALLOW_CUSTOM_ROLES:
        raise ValidationError('Creating custom roles is disabled')

    if content_type is None and (not system_roles_enabled()):
        raise ValidationError({'content_type': 'System-wide roles are not enabled'})

    if not settings.ANSIBLE_BASE_ALLOW_CUSTOM_TEAM_ROLES and content_type:
        if content_type.id == permission_registry.team_ct_id:
            raise ValidationError('Creating custom roles for teams is disabled')
        for perm in permissions:
            if perm.content_type_id == permission_registry.team_ct_id:
                raise ValidationError('Creating custom roles that include team permissions is disabled')


def check_view_permission_criteria(codename_set: set[str], permissions_by_model: dict[Type[Model], list[str]]) -> None:
    """Given a codename_set to be used in a role definition, enforce that view permission is included

    For example, a role can not give change permission to a thing without also giving view permission,
    because being able to change a thing without the ability to see it makes no sense.
    """
    for cls, valid_model_permissions in permissions_by_model.items():
        if 'view' in cls._meta.default_permissions:
            model_permissions = set(valid_model_permissions) & codename_set
            local_codenames = {codename for codename in model_permissions if not is_add_perm(codename)}
            if local_codenames and not any('view' in codename for codename in local_codenames):
                raise ValidationError(
                    {'permissions': f'Permissions for model {cls._meta.verbose_name} needs to include view, got: {prnt_codenames(local_codenames)}'}
                )


def check_has_change_with_delete(codename_set: set[str], permissions_by_model: dict[Type[Model], list[str]]):
    """Given a codename_set to be used in a role definition, include change if including delete

    We would like to get rid of this criteria eventually, but no harm in making it configurable.
    This requires that listing a delete permission in a RoleDefinition, like delete_credential,
    also includes the respective change permission like change_credential.
    If it has delete permission without change permission, throw an error.
    """
    for cls, valid_model_permissions in permissions_by_model.items():
        if 'delete' in cls._meta.default_permissions and 'change' in cls._meta.default_permissions:
            model_permissions = set(valid_model_permissions) & codename_set
            local_codenames = {codename for codename in model_permissions if not is_add_perm(codename)}
            if any('delete' in codename for codename in local_codenames) and not any('change' in codename for codename in local_codenames):
                raise ValidationError(
                    {'permissions': f'Permissions for model {cls._meta.verbose_name} needs to include change, got: {prnt_codenames(local_codenames)}'}
                )


def validate_permissions_for_model(permissions, content_type: Optional[Model], managed: bool = False) -> None:
    """Validation for creating a RoleDefinition

    This is called by the RoleDefinitionSerializer so clients will get these errors.
    It is also called by manager helper methods like RoleDefinition.objects.create_from_permissions
    which is done as an aid to tests and other apps integrating this library.
    """
    if not managed:
        validate_role_definition_enabled(permissions, content_type)

    codename_set = {perm.codename for perm in permissions}
    if content_type is None and permission_registry.team_permission in codename_set:
        # Special validation case, global team permissions are not allowed in any scenario
        raise ValidationError({'permissions': f'The {permission_registry.team_permission} permission can not be used in global roles'})

    role_model = None
    if content_type:
        role_model = content_type.model_class()
    permissions_by_model = permissions_allowed_for_role(role_model)

    invalid_codenames = codename_set - combine_values(permissions_by_model)
    if invalid_codenames:
        print_codenames = ', '.join(f'"{codename}"' for codename in invalid_codenames)
        raise ValidationError({'permissions': f'Permissions {print_codenames} are not valid for {prnt_model_name(role_model)} roles'})

    # Check that view permission is given for every model that has update/delete/special actions listed
    if settings.ANSIBLE_BASE_ROLES_REQUIRE_VIEW:
        check_view_permission_criteria(codename_set, permissions_by_model)

    # Check requires change and role_model is a system-wide role (None means it is), skip this validation.
    if settings.ANSIBLE_BASE_DELETE_REQUIRE_CHANGE and role_model is not None:
        check_has_change_with_delete(codename_set, permissions_by_model)

    if (not managed) and (not settings.ALLOW_SHARED_RESOURCE_CUSTOM_ROLES):
        for perm in permissions:
            # View permission for shared objects is interpreted as permission to view
            # the resource locally, which is needed to be able to view parent objects
            # For system roles, this exception is not allowed
            if content_type and perm.codename.startswith('view'):
                continue
            model = perm.content_type.model_class()
            if permission_registry.get_resource_prefix(model) == 'shared':
                raise ValidationError({'permissions', 'Local custom roles can only include view permission for shared models'})


def validate_codename_for_model(codename: str, model: Union[Model, Type[Model]]) -> str:
    """Shortcut method and validation to allow action name, codename, or app_name.codename

    This institutes a shortcut for easier use of the evaluation methods
    so that user.has_obj_perm(obj, 'change') is the same as user.has_obj_perm(obj, 'change_inventory')
    assuming obj is an inventory.
    It also tries to protect the user by throwing an error if the permission does not work.
    """
    valid_codenames = codenames_for_cls(model)
    if (not codename.startswith('add')) and codename in valid_codenames:
        return codename
    if re.match(r'^[a-z]+$', codename):
        # convience to call JobTemplate.accessible_objects(u, 'execute')
        name = f'{codename}_{model._meta.model_name}'
    else:
        # sometimes permissions are referred to with the app name, like test_app.say_cow
        if '.' in codename:
            name = codename.split('.')[-1]
        else:
            name = codename
    if name in valid_codenames:
        if name.startswith('add'):
            raise RuntimeError(f'Add permissions only valid for parent models, received for {model._meta.model_name}')
        return name

    for rel, child_cls in permission_registry.get_child_models(model):
        if name in codenames_for_cls(child_cls):
            return name
    raise RuntimeError(f'The permission {name} is not valid for model {model._meta.model_name}')


def validate_team_assignment_enabled(content_type: Optional[Model], has_team_perm: bool = False, has_org_member: bool = False) -> None:
    """Called in role assignment logic, inside RoleDefinition.give_permission

    Raises error if a setting disables the kind of permission being given.
    This mostly deals with team permissions.
    """
    team_team_allowed = settings.ANSIBLE_BASE_ALLOW_TEAM_PARENTS
    team_org_allowed = settings.ANSIBLE_BASE_ALLOW_TEAM_ORG_PERMS
    team_org_member_allowed = settings.ANSIBLE_BASE_ALLOW_TEAM_ORG_MEMBER
    team_org_team_allowed = settings.ANSIBLE_BASE_ALLOW_TEAM_ORG_ADMIN

    if all([team_team_allowed, team_org_allowed, team_org_member_allowed, team_org_team_allowed]):
        return  # Everything is allowed
    team_model_name = permission_registry.team_model._meta.model_name

    if not team_team_allowed and content_type.model == team_model_name:
        raise ValidationError('Assigning team permissions to other teams is not allowed')

    team_parent_model = permission_registry.get_parent_model(permission_registry.team_model)
    if team_parent_model is not None:
        team_parent_model_name = team_parent_model._meta.model_name
        if not team_org_allowed and content_type.model == team_parent_model_name:
            raise ValidationError(f'Assigning {team_parent_model_name} permissions to teams is not allowed')

        if (not team_org_member_allowed) and has_org_member:
            raise ValidationError(f'Assigning {team_parent_model_name} member permission to teams is not allowed')

        if not team_org_team_allowed and content_type.model == team_parent_model_name and has_team_perm:
            raise ValidationError(f'Assigning {team_parent_model_name} permissions that manage other teams is not allowed')


def validate_assignment(rd, actor, obj) -> None:
    """General validation for making a role assignment

    This is called programatically in the give_permission and give_global_permission methods.
    Some of this covered by serializers as well by basic field validation and param gathering.
    """
    if actor._meta.model_name not in ('user', 'team'):
        raise ValidationError(f'Cannot give permission to {actor}, must be a user or team')

    obj_ct = permission_registry.content_type_model.objects.get_for_model(obj)
    if obj_ct.id != rd.content_type_id:
        rd_model = getattr(rd.content_type, "model", "global")
        raise ValidationError(f'Role type {rd_model} does not match object {obj_ct.model}')


def check_locally_managed(rd: Model) -> None:
    """Can the given role definition be managed here, or is it externally managed

    This rule is a bridge solution until the RoleDefinition model declares
    explicitly whether it is managed locally or by a remote system.
    """
    if not (('ansible_base.jwt_consumer' in settings.INSTALLED_APPS) and (not settings.ALLOW_LOCAL_ASSIGNING_JWT_ROLES)):
        return
    if rd.name in settings.ANSIBLE_BASE_JWT_MANAGED_ROLES:
        raise ValidationError('Not managed locally, use the resource server instead')



import threading
from contextlib import contextmanager
from functools import lru_cache

from ansible_base.resource_registry.models import Resource, init_resource_from_object
from ansible_base.resource_registry.registry import get_registry
from ansible_base.resource_registry.utils.sync_to_resource_server import sync_to_resource_server


@lru_cache(maxsize=1)
def get_resource_models():
    resource_models = set()
    registry = get_registry()
    if registry:
        for k, resource in registry.get_resources().items():
            resource_models.add(resource.model)

    return resource_models


def remove_resource(sender, instance, **kwargs):
    try:
        resource = Resource.get_resource_for_object(instance)
        resource.delete()
    except Resource.DoesNotExist:
        return


def update_resource(sender, instance, created, **kwargs):
    try:
        resource = Resource.get_resource_for_object(instance)
        resource.update_from_content_object()
    except Resource.DoesNotExist:
        resource = init_resource_from_object(instance)
        resource.save()


# pre_save
def decide_to_sync_update(sender, instance, raw, using, update_fields, **kwargs):
    """
    A pre_save hook that decides whether or not to reverse-sync the instance
    based on which fields have changed.

    This has to be in pre-save because we have to be able to get the original
    instance to calculate which fields changed, if update_fields wasn't passed
    """

    if instance._state.adding:
        # We only concern ourselves with updates
        return

    try:
        resource = Resource.get_resource_for_object(instance)
    except Resource.DoesNotExist:
        # We can't sync here, but we want to log that, so let sync_to_resource_server() discard it.
        return

    fields_that_sync = resource.content_type.resource_type.serializer_class().get_fields().keys()

    if update_fields is None:
        # If we're not given a useful update_fields, manually calculate the changed fields
        # at the cost of an extra query
        existing_instance = sender.objects.get(pk=instance.pk)
        changed_fields = set()
        for field in fields_that_sync:
            if getattr(existing_instance, field) != getattr(instance, field):
                changed_fields.add(field)
    else:
        # If we're given update_fields, we can just check those
        changed_fields = set(update_fields)

    if not changed_fields.intersection(fields_that_sync):
        instance._skip_reverse_resource_sync = True


class ReverseSyncEnabled(threading.local):
    def __init__(self):
        self.enabled = True

    def __bool__(self):
        return self.enabled


reverse_sync_enabled = ReverseSyncEnabled()


@contextmanager
def no_reverse_sync():
    previous_value = reverse_sync_enabled.enabled
    reverse_sync_enabled.enabled = False
    try:
        yield
    finally:
        reverse_sync_enabled.enabled = previous_value


# post_save
def sync_to_resource_server_post_save(sender, instance, created, update_fields, **kwargs):
    if not reverse_sync_enabled:
        return

    action = "create" if created else "update"
    sync_to_resource_server(instance, action)


# pre_delete
def sync_to_resource_server_pre_delete(sender, instance, **kwargs):
    if not reverse_sync_enabled:
        return

    sync_to_resource_server(instance, "delete", ansible_id=instance.resource.ansible_id)



import logging
from collections import defaultdict
from typing import Optional
from uuid import UUID

from django.conf import settings

from ansible_base.rbac.models import ObjectRole, RoleDefinition, RoleEvaluation, RoleEvaluationUUID
from ansible_base.rbac.permission_registry import permission_registry
from ansible_base.rbac.prefetch import TypesPrefetch

logger = logging.getLogger('ansible_base.rbac.caching')


"""
This module has callable methods to fill in things marked with COMPUTED DATA in the models
from the user-specifications in other fields.
These need to be called in specific hooks to assure that evaluations remain correct,
logic for triggers are in the triggers module.

NOTE:
This is highly dependent on the model methods ObjectRole.needed_cache_updates and expected_direct_permissions
Those methods are what truly dictate the object-role to object-permission translation
"""


def all_team_parents(team_id: int, team_team_parents: dict, seen: Optional[set] = None) -> set[int]:
    """
    Returns parent teams, and parent teams of parent teams, until we have them all
        {parent_team_id, parent_team_id, ...}

    team_id: id of the team we want to get the direct and indirect parents of
    team_team_parents: mapping of team id to ids of its parents, this is not modified by this method
    seen: mutable set that will be added to by each call so that we can not recurse infinitely
    """
    parent_team_ids = set()
    if seen is None:
        seen = set()
    for parent_id in team_team_parents.get(team_id, []):
        if parent_id in seen:
            # will be combined in a lower level of the call stack
            # this condition prevents infinite recursion in the event of loops in the graph
            continue
        parent_team_ids.add(parent_id)
        seen.add(parent_id)
        parent_team_ids.update(all_team_parents(parent_id, team_team_parents, seen=seen))
    return parent_team_ids


def get_org_team_mapping() -> dict[int, list[int]]:
    """
    Returns the teams in all organization as a dictionary.
        {
            organization_id: [team_id, team_id, ...],
            organization_id: [team_id, ...]
        }
    """
    org_team_mapping = defaultdict(list)
    team_fields = ['id']
    team_parent_fd = permission_registry.get_parent_fd_name(permission_registry.team_model)
    if team_parent_fd:
        team_fields.append(f'{team_parent_fd}_id')
        for team in permission_registry.team_model.objects.only(*team_fields):
            team_parent_id = getattr(team, f'{team_parent_fd}_id')
            org_team_mapping[team_parent_id].append(team.id)
    return org_team_mapping


def get_direct_team_member_roles(org_team_mapping: dict) -> dict[int, list[int]]:
    """
    If an organization-level role lists "member_team" permission, that confers
    several team's permissions to users who holds an org role of that type.
    If a team-level role lists "member_team" then that also convers
    the member permissions to the user.
    These do not yet consider teams-of-teams, so these are "direct" membership roles to a team.
    Returns a dictionary with teams as keys and the object role ids that give membership as values.
        {
            team_id: [role_id, role_id, ...],
            team_id: [role_id, ...]
        }
    """
    direct_member_roles = defaultdict(list)
    for object_role in ObjectRole.objects.filter(role_definition__permissions__codename=permission_registry.team_permission).iterator():
        if object_role.content_type_id == permission_registry.team_ct_id:
            direct_member_roles[int(object_role.object_id)].append(object_role.id)
        elif object_role.content_type_id == permission_registry.org_ct_id:
            object_id = int(object_role.object_id)
            if object_id not in org_team_mapping:
                continue  # this means the organization has no team but has member_team as a listed permission
            for team_id in org_team_mapping[object_id]:
                direct_member_roles[team_id].append(object_role.id)
        else:
            logger.warning(f'{object_role} gives {permission_registry.team_permission} to an invalid type')
    return direct_member_roles


def get_parent_teams_of_teams(org_team_mapping: dict) -> dict[int, list[int]]:
    """
    Returns a dictionary showing the teams-of-teams relationships in the system
    this happens when a member_team role confers membership to another team.
        {
            team_id: [parent_team_id, parent_team_id, ...],
            team_id: []
        }
    The queryset and logic is similar to get_direct_team_member_roles but
    optimizations are different.
    """
    team_team_parents = defaultdict(list)
    for object_role in ObjectRole.objects.filter(
        role_definition__permissions__codename=permission_registry.team_permission, teams__isnull=False
    ).prefetch_related('teams'):
        for actor_team in object_role.teams.all():
            if object_role.content_type_id == permission_registry.team_ct_id:
                team_team_parents[int(object_role.object_id)].append(actor_team.id)
            elif object_role.content_type_id == permission_registry.org_ct_id:
                object_id = int(object_role.object_id)
                if object_id not in org_team_mapping:
                    continue  # again, means the organization has no team but has member_team as a listed permission
                for team_id in org_team_mapping[object_id]:
                    team_team_parents[team_id].append(actor_team.id)
    return team_team_parents


def compute_team_member_roles():
    """
    Fills in the ObjectRole.provides_teams relationship for all teams.
    This relationship is a list of teams that the role grants membership for
    This method is always ran globally.
    """
    # Manually prefetch the team to org memberships
    org_team_mapping = get_org_team_mapping()

    # Build out the direct member roles for teams
    direct_member_roles = get_direct_team_member_roles(org_team_mapping)

    # Build a team-to-team child-to-parents mapping for teams that have permission to other teams
    team_team_parents = get_parent_teams_of_teams(org_team_mapping)

    # Now we need to crawl the team-team graph to get the full list of roles that grants access to each team
    # for each parent team that grants membership to a team, we need to add the roles that grant
    # membership to that parent team
    all_member_roles = {}
    for team_id, member_roles in direct_member_roles.items():
        all_member_roles[team_id] = set(member_roles)  # will also avoid mutating original data structure later
        for parent_team_id in all_team_parents(team_id, team_team_parents):
            all_member_roles[team_id].update(set(direct_member_roles.get(parent_team_id, [])))

    # Great! we should be done building all_member_roles which tells what roles gives team membership for all teams
    # now at this point we save that data
    for team in permission_registry.team_model.objects.prefetch_related('member_roles'):
        # NOTE: the .set method will not use the prefetched data, thus the messy implementation here
        existing_ids = set(r.id for r in team.member_roles.all())
        expected_ids = set(all_member_roles.get(team.id, []))
        to_add = expected_ids - existing_ids
        to_remove = existing_ids - expected_ids
        if to_add:
            team.member_roles.add(*to_add)
        if to_remove:
            team.member_roles.remove(*to_remove)


def compute_object_role_permissions(object_roles=None, types_prefetch=None):
    """
    Assumes the ObjectRole.provides_teams relationship is correct.
    Makes the RoleEvaluation table correct for all specified object_roles
    """
    to_delete = set()
    to_add = []

    if types_prefetch is None:
        types_prefetch = TypesPrefetch.from_database(RoleDefinition)
    if object_roles is None:
        object_roles = ObjectRole.objects.iterator()

    for object_role in object_roles:
        role_to_delete, role_to_add = object_role.needed_cache_updates(types_prefetch=types_prefetch)

        if role_to_delete:
            logger.debug(f'Removing {len(role_to_delete)} object-permissions from {object_role}')
            to_delete.update(role_to_delete)

        if role_to_add:
            logger.debug(f'Adding {len(role_to_add)} object-permissions to {object_role}')
            to_add.extend(role_to_add)

    if to_add:
        logger.info(f'Adding {len(to_add)} object-permission records')
        to_add_int = []
        to_add_uuid = []
        for evaluation in to_add:
            if isinstance(evaluation.object_id, int):
                to_add_int.append(evaluation)
            elif isinstance(evaluation.object_id, UUID):
                to_add_uuid.append(evaluation)
            else:
                raise RuntimeError(f'Could not find a place in cache for {evaluation}')
        if to_add_int:
            RoleEvaluation.objects.bulk_create(to_add_int, ignore_conflicts=settings.ANSIBLE_BASE_EVALUATIONS_IGNORE_CONFLICTS)
        if to_add_uuid:
            RoleEvaluationUUID.objects.bulk_create(to_add_uuid, ignore_conflicts=settings.ANSIBLE_BASE_EVALUATIONS_IGNORE_CONFLICTS)

    if to_delete:
        logger.info(f'Deleting {len(to_delete)} object-permission records')
        to_delete_int = []
        to_delete_uuid = []
        for evaluation_id, evaluation_type in to_delete:
            if evaluation_type is int:
                to_delete_int.append(evaluation_id)
            elif evaluation_type is UUID:
                to_delete_uuid.append(evaluation_id)
            else:
                raise RuntimeError(f'Unexpected type to delete {evaluation_id}-{evaluation_type}')
        if to_delete_int:
            RoleEvaluation.objects.filter(id__in=to_delete_int).delete()
        if to_delete_uuid:
            RoleEvaluationUUID.objects.filter(id__in=to_delete_uuid).delete()



import logging

from drf_spectacular.extensions import OpenApiAuthenticationExtension

from ansible_base.jwt_consumer.common.auth import JWTAuthentication

logger = logging.getLogger("ansible_base.jwt_consumer.eda.auth")


class EDAJWTAuthentication(JWTAuthentication):
    use_rbac_permissions = True


class EDAJWTAuthScheme(OpenApiAuthenticationExtension):
    target_class = EDAJWTAuthentication
    name = "EDAJWTAuthentication"

    def get_security_definition(self, auto_schema):
        return {"type": "apiKey", "name": "X-DAB-JW-TOKEN", "in": "header"}



from django.db.models import SET_NULL, ForeignKey, JSONField, fields

from ansible_base.authentication.authenticator_plugins.utils import generate_authenticator_slug, get_authenticator_plugin
from ansible_base.lib.abstract_models.common import UniqueNamedCommonModel
from ansible_base.lib.utils.models import prevent_search


def get_next_authenticator_order():
    """
    Returns the next authenticator order, which is equals to max(order) + 1
    """
    largest_order_authenticator = Authenticator.objects.values('order').order_by('-order').first()
    return largest_order_authenticator['order'] + 1 if largest_order_authenticator else 1


class Authenticator(UniqueNamedCommonModel):
    ignore_relations = ['authenticator_users']
    enabled = fields.BooleanField(default=False, help_text="Should this authenticator be enabled.")
    create_objects = fields.BooleanField(default=True, help_text="Allow authenticator to create objects (users, teams, organizations).")
    remove_users = fields.BooleanField(
        default=True, help_text="When a user authenticates from this source should they be removed from any other groups they were previously added to."
    )
    configuration = prevent_search(JSONField(default=dict, help_text="The required configuration for this source.", blank=True))
    type = fields.CharField(
        editable=False,
        max_length=256,
        help_text="The type of authentication service this is.",
    )
    order = fields.IntegerField(
        default=get_next_authenticator_order,
        help_text="The order in which an authenticator will be tried. This only pertains to username/password authenticators.",
    )
    slug = fields.SlugField(
        max_length=1024,
        default=None,
        editable=False,
        unique=True,
        help_text="An immutable identifier for the authenticator; used to generate the sso uri for sso authenticator types",
    )
    category = fields.CharField(max_length=30, default=None, help_text="The base type of this authenticator.")

    auto_migrate_users_to = ForeignKey(
        "Authenticator",
        help_text=(
            "Automatically move users from this authenticator to the target authenticator when a matching user logs in via the target authenticator. "
            "For this to work, the field used for the user ID on both authenticators needs to have the same value. This should only be used when "
            "migrating users between two authentication mechanisms that share the same user database (such as when both IDPs share the same "
            "LDAP user directory)."
        ),
        related_name="auto_migrate_users_from",
        null=True,
        on_delete=SET_NULL,
    )

    def save(self, *args, **kwargs):
        from ansible_base.lib.utils.encryption import ansible_encryption

        # Here we are going to allow an exception to raise because what else can we do at this point?
        authenticator = get_authenticator_plugin(self.type)

        if not self.category:
            self.category = authenticator.category

        for field in getattr(authenticator, 'configuration_encrypted_fields', []):
            if field in self.configuration:
                self.configuration[field] = ansible_encryption.encrypt_string(self.configuration[field])

        if not self.slug:
            self.slug = generate_authenticator_slug()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.name

    @classmethod
    def from_db(cls, db, field_names, values):
        from ansible_base.lib.utils.encryption import ansible_encryption

        instance = super().from_db(db, field_names, values)

        try:
            authenticator = get_authenticator_plugin(instance.type)
            for field in getattr(authenticator, 'configuration_encrypted_fields', []):
                if field in instance.configuration:
                    instance.configuration[field] = ansible_encryption.decrypt_string(instance.configuration[field])
        except ImportError:
            # A log message will already be displayed if this fails
            pass

        return instance

    def get_login_url(self):
        plugin = get_authenticator_plugin(self.type)
        return plugin.get_login_url(self)

    def related_fields(self, request):
        response = super().related_fields(request)

        try:
            plugin = get_authenticator_plugin(self.type)
            response.update(plugin.add_related_fields(request, self))
        except ImportError:
            # If the plugin was removed we could get an ImportError but we still want to return what we can.
            pass

        return response

    @property
    def is_last_enabled(self):
        """
        Is this the -only- "enabled" authenticator?

        If used in a conditional statement, this property provides the combination
        of two conditions:
            A) is this enabled?
            B) is this the only authenticator enabled?

        Use this to gate deletion or disabling of authenticators so that the system
        will always have at least one enabled.

        Given that, if the authenticator is NOT enabled, this will always return false
        and could be misleading if no other authenticators are enabled. If you need to
        check if any authenticator is enabled, you should explicitly check with a
        queryset including a filter and an exists call. Do not rely on this property
        alone when making decisions about what can be deleted or disabled.
        """
        if self.enabled and not self.__class__.objects.filter(enabled=True).exclude(id=self.id).exists():
            return True
        return False



import hashlib

import oauth2_provider.models as oauth2_models
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.abstract_models.common import CommonModel
from ansible_base.lib.utils.hashing import hash_string
from ansible_base.lib.utils.models import prevent_search

activitystream = object
if 'ansible_base.activitystream' in settings.INSTALLED_APPS:
    from ansible_base.activitystream.models import AuditableModel

    activitystream = AuditableModel


class OAuth2RefreshToken(CommonModel, oauth2_models.AbstractRefreshToken, activitystream):
    class Meta(oauth2_models.AbstractRefreshToken.Meta):
        verbose_name = _('refresh token')
        ordering = ('id',)
        swappable = "OAUTH2_PROVIDER_REFRESH_TOKEN_MODEL"

    token = prevent_search(models.CharField(max_length=255, help_text=_("The refresh token value.")))
    updated = None  # Tracked in CommonModel with 'modified', no need for this

    def save(self, *args, **kwargs):
        if not self.pk:
            self.token = hash_string(self.token, hasher=hashlib.sha256, algo="sha256")
        super().save(*args, **kwargs)



import logging

from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from tacacs_plus.client import TACACSClient
from tacacs_plus.flags import TAC_PLUS_AUTHEN_TYPES, TAC_PLUS_VIRTUAL_REM_ADDR

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.social_auth import SocialAuthMixin
from ansible_base.authentication.utils.authentication import get_or_create_authenticator_user
from ansible_base.authentication.utils.claims import update_user_claims
from ansible_base.lib.serializers.fields import BooleanField, CharField, ChoiceField, IntegerField
from ansible_base.lib.utils.requests import get_remote_host

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.tacacs')

#
# Someone created a django library for tacacs: https://github.com/mwallraf/django-auth-tacacs
# It's appears to leverage tacacs-plus but has a single maintainer and has not been touched in a year.
# There are no issues on the repo but there has only been one release.
# The library is farily minimal so for now we are just going to do our own coding in this authenticator.
# Later on, if we choose, it should be a simple lift to use django-auth-tacacs
#


def validate_tacacsplus_disallow_nonascii(value):
    try:
        value.encode('ascii')
    except (UnicodeEncodeError, UnicodeDecodeError, AttributeError):
        raise ValidationError(_('TACACS+ secret does not allow non-ascii characters'))


class TacacsConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://github.com/ansible/tacacs_plus"
    HOST = CharField(
        label="TACACS+ Server",
        required=True,
        help_text="Hostname of TACACS+ server.",
        ui_field_label=_('Hostname of TACACS+ Server'),
    )
    PORT = IntegerField(
        min_value=1,
        max_value=65535,
        default=49,
        label=_('TACACS+ Port'),
        help_text=_('Port number of TACACS+ server.'),
        ui_field_label=_('Port number of TACACS+ Server'),
    )
    AUTH_PROTOCOL = ChoiceField(
        choices=['ascii', 'pap', 'chap'],
        default='ascii',
        label=_('TACACS+ Authentication Protocol'),
        help_text=_('Choose the authentication protocol used by TACACS+ client.'),
        ui_field_label=_('TACACS+ Authentication Protocol'),
    )
    REM_ADDR = BooleanField(
        default=True,
        label=_('TACACS+ client address sending enabled'),
        help_text=_('Enable the client address sending by TACACS+ client.'),
        ui_field_label=_('TACACS+ client address sending enabled'),
    )
    SECRET = CharField(
        validators=[validate_tacacsplus_disallow_nonascii],
        label=_('TACACS+ Secret'),
        help_text=_('Shared secret for authenticating to TACACS+ server.'),
        ui_field_label=_('Shared secret for authenticating to TACACS+ server.'),
        required=True,
    )
    SESSION_TIMEOUT = IntegerField(
        min_value=0,
        default=5,
        label=_('TACACS+ Auth Session Timeout'),
        help_text=_('TACACS+ session timeout value in seconds, 0 disables timeout.'),
        ui_field_label=_('TACACS+ Auth Session Timeout'),
    )


class AuthenticatorPlugin(SocialAuthMixin, AbstractAuthenticatorPlugin, ModelBackend):
    configuration_class = TacacsConfiguration
    logger = logger
    type = "tacacs"
    category = "password"

    def __init__(self, database_instance=None, *args, **kwargs):
        super().__init__(database_instance, *args, **kwargs)
        self.configuration_encrypted_fields = ['SECRET']

    def authenticate(self, request, username=None, password=None, **kwargs):
        if not username or not password:
            return None

        if not self.database_instance:
            logger.error("AuthenticatorPlugin was missing an authenticator")
            return None

        if not self.database_instance.enabled:
            logger.info(f"TACACS authenticator {self.database_instance.name} is disabled, skipping")
            return None

        try:
            tacacs_client = TACACSClient(
                self.database_instance.configuration["HOST"],
                self.database_instance.configuration["PORT"],
                self.database_instance.configuration["SECRET"],
                timeout=self.database_instance.configuration["SESSION_TIMEOUT"],
            )
            rem_addr = TAC_PLUS_VIRTUAL_REM_ADDR
            client_ip = get_remote_host(request)
            if client_ip:
                rem_addr = client_ip

            reply = tacacs_client.authenticate(
                username,
                password,
                authen_type=TAC_PLUS_AUTHEN_TYPES[self.database_instance.configuration['AUTH_PROTOCOL']],
                rem_addr=rem_addr,
            )

            if reply.valid:
                user, _authenticator_user, _created = get_or_create_authenticator_user(
                    username,
                    self.database_instance,
                    user_details={},
                    extra_data={'username': username},
                )

                return update_user_claims(user, self.database_instance, [])
        except Exception as e:
            logger.exception("TACACS+ Authentication Error: %s" % str(e))

        # Tacacs could not validate us so return None.
        return None



import inspect
import logging
import re
from collections import OrderedDict
from typing import Any

import ldap
from django.utils.translation import gettext_lazy as _
from django_auth_ldap import config
from django_auth_ldap.backend import LDAPBackend
from django_auth_ldap.backend import LDAPSettings as BaseLDAPSettings
from django_auth_ldap.config import LDAPGroupType
from rest_framework.serializers import ValidationError

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, Authenticator, BaseAuthenticatorConfiguration
from ansible_base.authentication.utils.authentication import get_or_create_authenticator_user
from ansible_base.authentication.utils.claims import update_user_claims
from ansible_base.lib.serializers.fields import BooleanField, CharField, ChoiceField, DictField, ListField, URLListField, UserAttrMap
from ansible_base.lib.utils.validation import VALID_STRING

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.ldap')


user_search_string = '%(user)s'


def validate_ldap_dn(value: str, with_user: bool = False, required: bool = True) -> None:
    if not value and not required:
        return

    dn_value = value
    if with_user:
        if user_search_string not in value:
            raise ValidationError(_('DN must include "{}" placeholder for username: {}').format(user_search_string, value))

        dn_value = value.replace(user_search_string, 'USER')

    try:
        ldap.dn.str2dn(dn_value.encode('utf-8'))
    except ldap.DECODING_ERROR:
        raise ValidationError(_('Invalid DN: %s') % value)


class DNField(CharField):
    def __init__(self, **kwargs):
        self.with_user = kwargs.pop('with_user', False)
        super().__init__(**kwargs)

        def validator(value):
            validate_ldap_dn(value, with_user=self.with_user, required=self.required)

        self.validators.append(validator)


class LDAPConnectionOptions(DictField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        def validator(value):
            valid_options = dict([(v, k) for k, v in ldap.OPT_NAMES_DICT.items()])
            errors = {}
            for key in value.keys():
                if key not in valid_options:
                    errors[key] = 'Not a valid connection option'
            if errors:
                raise ValidationError(errors)

        self.validators.append(validator)


class LDAPSearchField(ListField):
    def __init__(self, **kwargs):
        self.search_must_have_user = kwargs.pop('search_must_have_user', False)
        super().__init__(**kwargs)

        def validator(value):
            if not value:
                return

            errors = {}

            if len(value) != 3:
                raise ValidationError(_('Must be an array of 3 items: search DN, search scope and a filter'))

            try:
                validate_ldap_dn(value[0], with_user=False, required=True)
            except ValidationError as e:
                errors[0] = e.args[0]

            if type(value[1]) is not str or not value[1].startswith('SCOPE_') or not getattr(ldap, value[1], None):
                errors[1] = _('Must be a string representing an LDAP scope object')

            try:
                validate_ldap_filter(value[2], with_user=self.search_must_have_user)
            except ValidationError as e:
                errors[2] = e.args[0]

            if errors:
                raise ValidationError(errors)

            # We made it all the way here, make sure we can instantiate an LDAPSearch object
            try:
                # Search fields should be LDAPSearch objects, so we need to convert them from [] to these objects
                config.LDAPSearch(value[0], getattr(ldap, value[1]), value[2])
            except Exception as e:
                raise ValidationError(_('Failed to instantiate LDAPSearch object: %(e)s') % {"e": e})

        self.validators.append(validator)


def validate_ldap_filter(value: Any, with_user: bool = False) -> None:
    if type(value) is not str:
        raise ValidationError(VALID_STRING)

    value = value.strip()

    dn_value = value
    if with_user:
        if user_search_string not in value:
            raise ValidationError(_('DN must include "{}" placeholder for username: {}').format(user_search_string, value))
        dn_value = value.replace(user_search_string, 'USER')

    if re.match(r'^\([A-Za-z0-9-]+?=[^()]+?\)$', dn_value):
        return
    elif re.match(r'^\([&|!]\(.*?\)\)$', dn_value):
        for sub_filter in dn_value[3:-2].split(')('):
            # We only need to check with_user at the top of the recursion stack
            validate_ldap_filter(f'({sub_filter})', with_user=False)
        return
    raise ValidationError(_('Invalid filter: %s') % value)


def get_all_sub_classes(cls):
    # This function can get the names of all subclasses... maybe we want to move this into utils
    # We use it to find all of the parent classes for LDAPGroup
    sub_classes = []
    for sub_cls in cls.__subclasses__():
        sub_classes.append(sub_cls.__name__)
        sub_classes.extend(get_all_sub_classes(sub_cls))
    return sub_classes


class LDAPConfiguration(BaseAuthenticatorConfiguration):
    # We add group type params to our list of valid settings
    defaults = dict(list(BaseLDAPSettings.defaults.items()) + list({'GROUP_TYPE_PARAMS': {}}.items()))

    documentation_url = "https://django-auth-ldap.readthedocs.io/en/latest/"

    SERVER_URI = URLListField(
        help_text=_('A list of URIs to connect to LDAP server, such as "ldap://ldap.example.com:389" (non-SSL) or "ldaps://ldap.example.com:636" (SSL).'),
        allow_null=False,
        required=True,
        schemes=['ldap', 'ldaps'],
        ui_field_label=_('LDAP Server URI'),
    )

    BIND_DN = DNField(
        help_text=_(
            'DN (Distinguished Name) of user to bind for all search queries. This'
            ' is the system user account we will use to login to query LDAP for other'
            ' user information. Refer to the documentation for example syntax.'
        ),
        allow_null=False,
        required=False,
        with_user=False,
        ui_field_label=_('LDAP Bind DN'),
    )
    BIND_PASSWORD = CharField(
        help_text=_("The password used for BIND_DN."),
        allow_null=False,
        required=False,
        ui_field_label=_('LDAP Bind Password'),
    )

    CONNECTION_OPTIONS = LDAPConnectionOptions(
        help_text=_(
            'Additional options to set for the LDAP connection.  LDAP '
            'referrals are disabled by default (to prevent certain LDAP '
            'queries from hanging with AD). Option names should be strings '
            '(e.g. "OPT_REFERRALS"). Refer to '
            'https://www.python-ldap.org/doc/html/ldap.html#options for '
            'possible options and values that can be set.'
        ),
        default={},
        allow_null=False,
        required=False,
        ui_field_label=_('LDAP Connection Options'),
    )

    GROUP_TYPE = ChoiceField(
        help_text=_(
            'The group type may need to be changed based on the type of the '
            'LDAP server.  Values are listed at: '
            'https://django-auth-ldap.readthedocs.io/en/stable/groups.html#types-of-groups'
        ),
        allow_null=False,
        required=True,
        choices=get_all_sub_classes(LDAPGroupType),
        ui_field_label=_('LDAP Group Type'),
    )
    GROUP_TYPE_PARAMS = DictField(
        help_text=_('Key value parameters to send the chosen group type init method.'),
        # There is no default here because it depends on your GROUP_TYPE
        allow_null=False,
        required=True,
        ui_field_label=_('LDAP Group Type Parameters'),
    )
    GROUP_SEARCH = LDAPSearchField(
        help_text=_(
            'Users are mapped to organizations based on their membership in LDAP'
            ' groups. This setting defines the LDAP search query to find groups. '
            'Unlike the user search, group search does not support LDAPSearchUnion.'
        ),
        allow_null=True,
        required=False,
        search_must_have_user=False,
        ui_field_label=_('LDAP Group Search'),
    )
    START_TLS = BooleanField(
        help_text=_("Whether to enable TLS when the LDAP connection is not using SSL."),
        allow_null=False,
        required=False,
        default=False,
        ui_field_label=_('LDAP Start TLS'),
    )
    USER_DN_TEMPLATE = DNField(
        help_text=_(
            'Alternative to user search, if user DNs are all of the same '
            'format. This approach is more efficient for user lookups than '
            'searching if it is usable in your organizational environment. If '
            'this setting has a value it will be used instead of '
            'AUTH_LDAP_USER_SEARCH.'
        ),
        allow_null=True,
        required=False,
        with_user=True,
        ui_field_label=_('LDAP User DN Template'),
    )
    USER_ATTR_MAP = UserAttrMap(
        help_text=_(
            'Mapping of LDAP user schema to API user attributes. The default'
            ' setting is valid for ActiveDirectory but users with other LDAP'
            ' configurations may need to change the values. Refer to the'
            ' documentation for additional details.'
        ),
        allow_null=False,
        required=True,
        ui_field_label=_('LDAP User Attribute Map'),
    )
    USER_SEARCH = LDAPSearchField(
        help_text=_(
            'LDAP search query to find users.  Any user that matches the given '
            'pattern will be able to login to the service.  The user should also be '
            'mapped into an organization (as defined in the '
            'AUTH_LDAP_ORGANIZATION_MAP setting).  If multiple search queries '
            'need to be supported use of "LDAPUnion" is possible. See '
            'the documentation for details.'
        ),
        allow_null=True,
        required=False,
        search_must_have_user=True,
        ui_field_label=_('LDAP User Search'),
    )

    def validate(self, attrs):
        # Check interdependent fields
        errors = {}

        group_type_class = getattr(config, attrs['GROUP_TYPE'], None)
        if group_type_class:
            group_type_params = attrs['GROUP_TYPE_PARAMS']
            logger.error(f"Validating group type params for {attrs['GROUP_TYPE']}")
            class_args = inspect.getfullargspec(group_type_class.__init__).args[1:]
            invalid_keys = set(group_type_params) - set(class_args)
            missing_keys = set(class_args) - set(group_type_params)
            if invalid_keys:
                invalid_keys = sorted(list(invalid_keys))
                for key in invalid_keys:
                    errors[f'GROUP_TYPE_PARAMS.{key}'] = "Invalid option for specified GROUP_TYPE"

            if missing_keys:
                missing_keys = sorted(list(missing_keys))
                for key in missing_keys:
                    errors[f'GROUP_TYPE_PARAMS.{key}'] = "Missing required field for GROUP_TYPE"

        if errors:
            raise ValidationError(errors)

        # Raise some warnings if specific fields were used
        # TODO: Figure out how to display these warnings on a successful save
        # for field in ['USER_FLAGS_BY_GROUP', 'DENY_GROUP', 'REQUIRE_GROUP']:
        #    if field in data:
        #        self.warnings[field] = "It would be better to use the authenticator field instead of setting this field in the LDAP adapter"

        return super().validate(attrs)


class LDAPSettings(BaseLDAPSettings):
    def __init__(self, prefix: str = 'AUTH_LDAP_', defaults: dict = {}):
        # This init method double checks the passed defaults while initializing a settings objects
        super(LDAPSettings, self).__init__(prefix, defaults)

        # SERVER_URI needs to be a string, not an array
        setattr(self, 'SERVER_URI', ','.join(defaults['SERVER_URI']))

        # Connection options need to be set as {"integer": "value"} but our configuration has {"friendly_name": "value"} so we need to convert them
        connection_options = defaults.get('CONNECTION_OPTIONS', {})
        valid_options = dict([(v, k) for k, v in ldap.OPT_NAMES_DICT.items()])
        internal_data = {}
        for key in connection_options:
            internal_data[valid_options[key]] = connection_options[key]

        # If a DB-backed setting is specified that wipes out the
        # OPT_NETWORK_TIMEOUT, fall back to a sane default
        if ldap.OPT_NETWORK_TIMEOUT not in internal_data:
            internal_data[ldap.OPT_NETWORK_TIMEOUT] = 30

        # when specifying `.set_option()` calls for TLS in python-ldap, the
        # *order* in which you invoke them *matters*, particularly in Python3,
        # where dictionary insertion order is persisted
        #
        # specifically, it is *critical* that `ldap.OPT_X_TLS_NEWCTX` be set *last*
        # this manual sorting puts `OPT_X_TLS_NEWCTX` *after* other TLS-related
        # options
        #
        # see: https://github.com/python-ldap/python-ldap/issues/55
        newctx_option = internal_data.pop(ldap.OPT_X_TLS_NEWCTX, None)
        internal_data = OrderedDict(internal_data)
        if newctx_option is not None:
            internal_data[ldap.OPT_X_TLS_NEWCTX] = newctx_option

        setattr(self, 'CONNECTION_OPTIONS', internal_data)

        # Group type needs to be an object instead of a String so instantiate it
        group_type_class = getattr(config, defaults['GROUP_TYPE'], None)
        setattr(self, 'GROUP_TYPE', group_type_class(**defaults['GROUP_TYPE_PARAMS']))


class AuthenticatorPlugin(LDAPBackend, AbstractAuthenticatorPlugin):
    configuration_class = LDAPConfiguration
    type = 'LDAP'
    category = "password"

    def __init__(self, database_instance=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.database_instance = database_instance
        if database_instance:
            self.settings = LDAPSettings(defaults=database_instance.configuration)
        self.configuration_encrypted_fields = ['BIND_PASSWORD']
        self.set_logger(logger)

    def authenticate(self, request, username=None, password=None, **kwargs) -> (object, dict, list):
        if not username or not password:
            return
        users_groups = []

        if not self.database_instance:
            logger.error("AuthenticatorPlugin was missing an authenticator")
            return None

        if not self.database_instance.enabled:
            logger.info(f"LDAP authenticator {self.database_instance.name} is disabled, skipping")
            return None

        # We don't have to check if settings is None because it can never happen, the parent object will always return something

        if self.settings.START_TLS and ldap.OPT_X_TLS_REQUIRE_CERT in self.settings.CONNECTION_OPTIONS:
            # with python-ldap, if you want to set connection-specific TLS
            # parameters, you must also specify OPT_X_TLS_NEWCTX = 0
            # see: https://stackoverflow.com/a/29722445
            # see: https://stackoverflow.com/a/38136255
            self.settings.CONNECTION_OPTIONS[ldap.OPT_X_TLS_NEWCTX] = 0

        # Ensure USER_SEARCH and GROUP_SEARCH are converted into a search object
        for field, search_must_have_user in [('GROUP_SEARCH', False), ('USER_SEARCH', True)]:
            data = getattr(self.settings, field, None)
            # Ignore None or empty (e.g., [])
            if not data:
                setattr(self.settings, field, None)
            elif not isinstance(data, config.LDAPSearch):
                try:
                    # Search fields should be LDAPSearch objects, so we need to convert them from [] to these objects
                    search_object = config.LDAPSearch(data[0], getattr(ldap, data[1]), data[2])
                    setattr(self.settings, field, search_object)
                except Exception as e:
                    logger.error(f'Failed to instantiate {field} LDAPSearch object: {e}')
                    return None

        try:
            user_from_ldap = super().authenticate(request, username, password)

            self.process_login_messages(user_from_ldap, username)

            # If we didn't get a user we can return None
            if user_from_ldap is None:
                return None

            if user_from_ldap is not None and user_from_ldap.ldap_user:
                if getattr(self.settings, 'GROUP_SEARCH'):
                    users_groups = list(user_from_ldap.ldap_user._get_groups().get_group_dns())

                # If we have an LDAP user and that user we found has an user_from_ldap internal object and that object has a bound connection
                # Then we can try and force an unbind to close the sticky connection
                if user_from_ldap.ldap_user._connection_bound:
                    logger.debug(f"Forcing LDAP connection to close for {self.database_instance.name}")
                    try:
                        user_from_ldap.ldap_user._connection.unbind_s()
                        user_from_ldap.ldap_user._connection_bound = False
                    except Exception:
                        logger.exception(
                            f"Got unexpected LDAP exception when forcing LDAP disconnect for user {user_from_ldap.username}, login will still proceed"
                        )

            # In unit testing there were cases where the function we are in was being called before get_or_build_user.
            # Its unclear if that was just a byproduct of mocking or a real scenario.
            # Since this call is idempotent we are just going to call it again to ensure the AuthenticatorUser is created for update_user_claims
            get_or_create_authenticator_user(username, self.database_instance, user_details={}, extra_data=user_from_ldap.ldap_user.attrs.data)
            return update_user_claims(user_from_ldap, self.database_instance, users_groups)
        except Exception:
            logger.exception(f"Encountered an error authenticating to LDAP {self.database_instance.name}")
            return None

    def process_login_messages(self, ldap_user, username: str) -> None:
        if ldap_user is None:
            logger.info(f"User {username} could not be authenticated by LDAP {self.database_instance.name}")

            # If our login failed and we have REQUIRE or DENY group we can't tell that the user is in that but we want inform the admin via a log as a hint
            if self.settings.REQUIRE_GROUP and self.settings.DENY_GROUP:
                logger.info("Hint: is user missing required group or in deny group?")
            elif self.settings.REQUIRE_GROUP:
                logger.info("Hint: is user missing required group?")
            elif self.settings.DENY_GROUP:
                logger.info("Hint: is user in deny group?")
        else:
            logger.info(f"User {username} authenticated by LDAP {self.database_instance.name}")

    def update_settings(self, database_authenticator: Authenticator) -> None:
        self.settings = LDAPSettings(defaults=database_authenticator.configuration)

    def get_or_build_user(self, username, ldap_user):
        """
        This gets called by _LDAPUser to create the user in the database.
        """
        user, _authenticator_user, created = get_or_create_authenticator_user(
            username,
            self.database_instance,
            user_details={
                "username": username,
            },
            extra_data=ldap_user.attrs.data,
        )

        return user, created



from collections import defaultdict

from django.core.management.base import BaseCommand, CommandError

from ansible_base.rbac import permission_registry
from ansible_base.rbac.models import ObjectRole, RoleDefinition


class Command(BaseCommand):
    help = "Runs bug checking sanity checks, gets scale metrics, and recommendations for Role Based Access Control"

    def check_role_definitions(self):
        rd_ct = RoleDefinition.objects.count()
        self.stdout.write(f'Inspecting {rd_ct} role definitions')
        self.stdout.write('  checking for minimum of view permission')
        indexed_rds = defaultdict(list)
        for rd in RoleDefinition.objects.prefetch_related('permissions'):
            perm_list = list(rd.permissions.values_list('codename', flat=True))
            if not any(p.startswith('view_') for p in perm_list):
                self.stdout.write(self.style.WARNING(f'Role definition {rd.name} does not list any view permissions and this is considered invalid'))
                self.has_issues = True
            perm_set = frozenset(perm_list)
            indexed_rds[perm_set].append(rd)

        self.stdout.write('  checking for duplicate role definitions')
        for perm_set, rd_list in indexed_rds.items():
            if len(rd_list) > 1:
                self.stdout.write(self.style.WARNING('Found duplicate role definitions with same permissions list:'))
                for rd in rd_list:
                    self.stdout.write(f'   {rd}')

        object_role_ct = ObjectRole.objects.count()
        self.stdout.write(f'Inspecting {object_role_ct} object roles')
        self.stdout.write('  checking for invalid permissions for model type')
        for role in ObjectRole.objects.prefetch_related('role_definition__permissions', 'content_type'):
            for permission in role.role_definition.permissions.all():
                if permission.content_type_id != role.content_type_id:
                    if permission.content_type.model_class() not in set(
                        cls for filter_path, cls in permission_registry.get_child_models(role.content_type.model_class())
                    ):
                        self.stdout.write(
                            self.style.WARNING(f'Object role {role} has permission {permission.codename} for an unlike content type {permission.content_type}')
                        )

    def check_object_roles(self):
        self.stdout.write('  checking for up-to-date role evaluations')
        for role in ObjectRole.objects.all():
            to_delete, to_add = role.needed_cache_updates()
            if to_delete or to_add:
                self.stdout.write(
                    self.style.WARNING(f'Object role {role} does not have up-to-date role evaluations cached, this can happen if someone bypasses signals')
                )

        self.stdout.write('  checking for missing content object')
        for role in ObjectRole.objects.all():
            if not role.content_object:
                self.stdout.write(self.style.WARNING(f'Object role {role} has been orphaned, indicating that post_delete signals are broken'))

    def handle(self, *args, **options):
        self.has_issues = False
        self.check_role_definitions()
        self.check_object_roles()
        if not self.has_issues:
            self.stdout.write(self.style.SUCCESS('No issues were found'))
        else:
            raise CommandError('Checks completed, some potential issues were noted')



import logging

from django.utils.translation import gettext_lazy as _
from rest_framework.response import Response
from rest_framework.serializers import ValidationError

from ansible_base.authentication.models import Authenticator
from ansible_base.lib.utils.settings import get_setting, is_aoc_instance
from ansible_base.lib.utils.validation import validate_image_data, validate_url
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView

logger = logging.getLogger('ansible_base.authentication.views.ui_auth')


class UIAuth(AnsibleBaseDjangoAppApiView):
    authentication_classes = []
    permission_classes = []

    def get(self, request, format=None):
        response = generate_ui_auth_data()

        return Response(response)


def generate_ui_auth_data():
    authenticators = Authenticator.objects.filter(enabled=True)
    response = {
        'passwords': [],
        'ssos': [],
        'show_login_form': False,
        'login_redirect_override': '',
        'custom_login_info': '',
        'custom_logo': '',
        'managed_cloud_install': False,
    }

    for authenticator in authenticators:
        if authenticator.category == 'password':
            response['passwords'].append(
                {
                    'name': authenticator.name,
                }
            )
            response["show_login_form"] = True
        elif authenticator.category == 'sso':
            try:
                response['ssos'].append({'name': authenticator.name, 'login_url': authenticator.get_login_url(), 'type': authenticator.type.split('.')[-1]})
                response["show_login_form"] = True
            except ImportError:
                logger.error(f"There is an enabled authenticator id {authenticator.id} whose plugin is not working {authenticator.type}")
        else:
            logger.error(f"Don't know how to handle authenticator of type {authenticator.type}")

    try:
        login_redirect_override = get_setting('LOGIN_REDIRECT_OVERRIDE', '')
        # ignore validation if login_redirect_override is None or empty string
        if login_redirect_override is not None and login_redirect_override != '':
            validate_url(url=login_redirect_override, allow_plain_hostname=True)
            response['login_redirect_override'] = login_redirect_override
    except ValidationError:
        logger.error('LOGIN_REDIRECT_OVERRIDE was set but was not a valid URL, ignoring')

    custom_login_info = get_setting('custom_login_info', '')
    if isinstance(custom_login_info, str):
        response['custom_login_info'] = custom_login_info
    else:
        logger.error("custom_login_info was not a string")
        raise ValidationError(_("custom_login_info was set but was not a valid string, ignoring"))

    try:
        custom_logo = get_setting('custom_logo', '')
        validate_image_data(custom_logo)
        response['custom_logo'] = custom_logo
    except ValidationError:
        logger.error("custom_logo was set but was not a valid image data, ignoring")

    # The cloud managed setting is not customizable outside of a conf file
    response['managed_cloud_install'] = is_aoc_instance()

    return response



import logging

from django.urls import include, path, re_path

from ansible_base.authentication import views
from ansible_base.authentication.apps import AuthenticationConfig
from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_plugins, get_authenticator_urls
from ansible_base.authentication.views.authenticator_users import get_authenticator_user_view
from ansible_base.lib.routers import AssociationResourceRouter

logger = logging.getLogger('ansible_base.authentication.urls')

app_name = AuthenticationConfig.label

list_actions = {'get': 'list', 'post': 'create'}
detail_actions = {'get': 'retrieve', 'put': 'update', 'patch': 'partial_update', 'delete': 'destroy'}
view_only_list = {'get': 'list'}

api_version_urls = []

# Load urls from authenticator plugins
for plugin_name in get_authenticator_plugins():
    plugin_urls = get_authenticator_urls(plugin_name)
    if plugin_urls:
        api_version_urls.extend(plugin_urls)
        logger.debug(f"Loaded URLS from {plugin_name}")


router = AssociationResourceRouter()
router.register(
    'authenticators',
    views.AuthenticatorViewSet,
    related_views={
        'authenticator_maps': (views.AuthenticatorMapViewSet, 'authenticator_maps'),
    },
)
router.register(
    'authenticator_maps',
    views.AuthenticatorMapViewSet,
    related_views={
        'authenticators': (views.AuthenticatorViewSet, 'authenticators'),
    },
)

# Add a custom related authenticator/X/users endpoint (not a standard FK relation on authenticators)
auth_user_view = get_authenticator_user_view()
if auth_user_view is not None:
    api_version_urls.extend(
        [
            re_path(
                '^authenticators/(?P<pk>[0-9]+)/users/$',
                auth_user_view.as_view(view_only_list),
                name='authenticator-users-list',
            )
        ]
    )

api_version_urls.extend(
    [
        # Authenticators
        path('', include(router.urls)),
        # Plugin List
        path('authenticator_plugins/', views.AuthenticatorPluginView.as_view(), name='authenticator_plugin-view'),
        # Trigger definition
        path('trigger_definition/', views.TriggerDefinitionView.as_view(), name='trigger_definition-view'),
        path('ui_auth/', views.UIAuth.as_view(), name='ui_auth-view'),
    ]
)


api_urls = [
    path('social/', include('social_django.urls', namespace='social')),
]



from django.utils.translation import gettext_lazy as _
from rest_framework.response import Response

from ansible_base.feature_flags.serializers import FeatureFlagSerializer
from ansible_base.lib.utils.views.ansible_base import AnsibleBaseView

from .utils import get_django_flags


class FeatureFlagsStateListView(AnsibleBaseView):
    """
    A view class for displaying feature flags
    """

    serializer_class = FeatureFlagSerializer
    filter_backends = []
    name = _('Feature Flags')
    http_method_names = ['get', 'head']

    def get(self, request, format=None):
        self.serializer = FeatureFlagSerializer()
        return Response(self.serializer.to_representation())

    def get_queryset(self):
        return get_django_flags()



# Generated by Django 4.2.11 on 2024-05-28 08:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0010_alter_authenticatoruser_unique_together'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticatormap',
            name='role',
            field=models.CharField(blank=True, default=None, help_text='The role this map will grant the authenticating user to the targeted object', max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='map_type',
            field=models.CharField(choices=[('allow', 'allow'), ('is_superuser', 'is_superuser'), ('role', 'role'), ('organization', 'organization'), ('team', 'team')], default='team', help_text='What does the map work on, a team, organization, a user flag or is this an allow rule', max_length=17),
        ),
    ]



import logging
from datetime import timedelta
from typing import Optional

from crum import get_current_user
from django.core.exceptions import ObjectDoesNotExist
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from oauthlib.common import generate_token
from oauthlib.oauth2 import AccessDeniedError
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.serializers import SerializerMethodField

from ansible_base.lib.serializers.common import CommonModelSerializer
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.lib.utils.settings import get_setting
from ansible_base.oauth2_provider.models import OAuth2AccessToken, OAuth2RefreshToken
from ansible_base.oauth2_provider.models.access_token import SCOPES

logger = logging.getLogger("ansible_base.oauth2_provider.serializers.token")


class OAuth2TokenSerializer(CommonModelSerializer):
    refresh_token = SerializerMethodField()

    unencrypted_token = None  # Only used in POST so we can return the token in the response
    unencrypted_refresh_token = None  # Only used in POST so we can return the refresh token in the response

    class Meta:
        model = OAuth2AccessToken
        fields = CommonModelSerializer.Meta.fields + [x.name for x in OAuth2AccessToken._meta.concrete_fields] + ['refresh_token']
        # The source_refresh_token and id_token are the concrete field but we change them to just token and refresh_token
        # We wrap these in a try for when we need to make the initial models
        try:
            fields.remove('source_refresh_token')
        except ValueError:
            pass
        try:
            fields.remove('id_token')
        except ValueError:
            pass
        read_only_fields = ('user', 'token', 'expires', 'refresh_token')
        extra_kwargs = {'scope': {'allow_null': False, 'required': False}, 'user': {'allow_null': False, 'required': True}}

    def to_representation(self, instance):
        request = self.context.get('request', None)
        ret = super().to_representation(instance)
        if request and request.method == 'POST':
            # If we're creating the token, show it. Otherwise, show the encrypted string.
            ret['token'] = self.unencrypted_token
        else:
            ret['token'] = ENCRYPTED_STRING
        return ret

    def get_refresh_token(self, obj) -> Optional[str]:
        request = self.context.get('request')
        try:
            if not obj.refresh_token:
                return None
            elif request and request.method == 'POST':
                return self.unencrypted_refresh_token
            else:
                return ENCRYPTED_STRING
        except ObjectDoesNotExist:
            return None

    def _is_valid_scope(self, value):
        if not value or (not isinstance(value, str)):
            return False
        words = value.split()
        for word in words:
            if words.count(word) > 1:
                return False  # do not allow duplicates
            if word not in SCOPES:
                return False
        return True

    def validate_scope(self, value):
        if not self._is_valid_scope(value):
            raise ValidationError(_('Must be a simple space-separated string with allowed scopes {}.').format(SCOPES))
        return value

    def create(self, validated_data):
        current_user = get_current_user()
        validated_data['token'] = generate_token()
        expires_delta = get_setting('OAUTH2_PROVIDER', {}).get('ACCESS_TOKEN_EXPIRE_SECONDS', 0)
        if expires_delta == 0:
            logger.warning("OAUTH2_PROVIDER.ACCESS_TOKEN_EXPIRE_SECONDS was set to 0, creating token that has already expired")
        validated_data['expires'] = now() + timedelta(seconds=expires_delta)
        validated_data['user'] = self.context['request'].user
        self.unencrypted_token = validated_data.get('token')  # Before it is hashed

        try:
            obj = super().create(validated_data)
        except AccessDeniedError as e:
            raise PermissionDenied(str(e))

        if obj.application and obj.application.user:
            obj.user = obj.application.user
        obj.save()
        if obj.application:
            self.unencrypted_refresh_token = generate_token()
            OAuth2RefreshToken.objects.create(
                user=current_user,
                token=self.unencrypted_refresh_token,
                application=obj.application,
                access_token=obj,
            )
        return obj



from collections import OrderedDict

from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ChoiceField, ValidationError

from ansible_base.authentication.authenticator_plugins.utils import generate_authenticator_slug, get_authenticator_plugin, get_authenticator_plugins
from ansible_base.authentication.models import Authenticator
from ansible_base.lib.serializers.common import NamedCommonModelSerializer
from ansible_base.lib.serializers.mixins import ImmutableFieldsMixin
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.lib.utils.response import get_relative_url


class AuthenticatorSerializer(NamedCommonModelSerializer, ImmutableFieldsMixin):
    type = ChoiceField(get_authenticator_plugins())

    def validate_type(self, value):
        if self.instance and self.instance.type != value:
            raise ValidationError(_("Cannot change authenticator type after it has been created."))
        return value

    def validate_slug(self, value):
        # Only need to perform creation validation for immutable fields;
        # read_only fields do not include the value in input to the serializer.
        value = generate_authenticator_slug(value)

        return value

    class Meta:
        model = Authenticator
        fields = NamedCommonModelSerializer.Meta.fields + [x.name for x in Authenticator._meta.concrete_fields]
        fields.remove("category")
        immutable_fields = ["slug"]
        extra_kwargs = {
            "slug": {
                "allow_blank": True,
                "required": False,
            }
        }

    # TODO: Do we need/want to delve into dicts and search their keys?
    def to_representation(self, authenticator):
        ret = super().to_representation(authenticator)
        configuration = authenticator.configuration
        masked_configuration = OrderedDict()

        try:
            authenticator_plugin = get_authenticator_plugin(authenticator.type)
            encrypted_keys = authenticator_plugin.configuration_encrypted_fields

            # If the authenticator configuration has a to_representation we need to respect it
            ret['configuration'] = authenticator_plugin.to_representation(authenticator)

            keys = list(configuration.keys())
            keys.sort()
            # Mask any keys in the encryption that should be masked
            for key in keys:
                if key in encrypted_keys:
                    masked_configuration[key] = ENCRYPTED_STRING
                else:
                    masked_configuration[key] = configuration[key]
            ret['configuration'] = masked_configuration
        except ImportError:
            # A log message will already be displayed if we can't load this
            ret['configuration'] = {}
            ret['error'] = _('Failed to load the plugin behind this authenticator, configuration hidden to protect secrets')
            return ret

        # Generate a sso login URL if this is an sso category
        login_url = authenticator.get_login_url()
        if login_url:
            ret['sso_login_url'] = login_url

        return ret

    def to_internal_value(self, data):
        parsed_data = super().to_internal_value(data)

        # Incase type was not passed in the data (like from a patch) we need to take it from the existing instance
        type = parsed_data.get('type', getattr(self.instance, 'type', None))

        # Here we will let a stack trace propagate because we can't convert this thing to an internal value and we likely don't want to save
        authenticator_plugin = get_authenticator_plugin(type)

        encrypted_keys = authenticator_plugin.configuration_encrypted_fields

        configuration = parsed_data.get('configuration', {})

        for key in encrypted_keys:
            if configuration.get(key, None) and self.instance and configuration.get(key, None) == ENCRYPTED_STRING:
                configuration[key] = self.instance.configuration.get(key)

        return parsed_data

    def validate(self, data) -> dict:
        # prevent disabling the last enabled authenticator
        if data.get('enabled') is False:
            if self.instance and self.instance.is_last_enabled:
                raise ValidationError(_("At least one authenticator must be enabled"))

        validator_type = data.get('type', None)
        # if we didn't have a type, try to get the type of the existing object (if we have one)
        if not validator_type and self.instance:
            validator_type = self.instance.type

        configuration = data.get('configuration', None)
        # Not having configuration is only valid for a PATCH
        request = self.context.get('request', None)
        if not request or (request.method != 'PATCH' and configuration is None):
            raise ValidationError(_("You must specify configuration for the authenticator"))

        if auto_migrate := data.get('auto_migrate_users_to'):
            if auto_migrate.auto_migrate_users_to is not None:
                raise ValidationError(
                    {
                        "auto_migrate_users_to": _(
                            "The authenticator you have picked is already configured to auto migrate users from a different authenticator."
                        )
                    }
                )

        try:
            invalid_encrypted_keys = {}
            authenticator = get_authenticator_plugin(validator_type)
            data = authenticator.validate(self, data)

            if configuration or configuration == {}:
                for key in authenticator.configuration_encrypted_fields:
                    if not self.instance and configuration.get(key, None) == ENCRYPTED_STRING:
                        invalid_encrypted_keys[key] = _("Can not be set to %(ENCRYPTED_STRING)s") % {"ENCRYPTED_STRING": ENCRYPTED_STRING}
                if invalid_encrypted_keys:
                    raise ValidationError(invalid_encrypted_keys)
                data['configuration'] = authenticator.validate_configuration(configuration, self.instance)
            return data
        except ImportError as e:
            raise ValidationError({'type': _('Failed to import %(e)s') % {'e': e}})

    def _get_related(self, obj) -> dict[str, str]:
        related = super()._get_related(obj)
        from ansible_base.authentication.views.authenticator_users import get_authenticator_user_view

        if get_authenticator_user_view():
            related['users'] = get_relative_url('authenticator-users-list', kwargs={'pk': obj.pk})

        return related



from __future__ import annotations

import inspect
import os
import sys
from pathlib import Path
from typing import Any

from django.core.exceptions import ImproperlyConfigured
from dynaconf import Dynaconf, Validator
from dynaconf.constants import YAML_EXTENSIONS
from dynaconf.loaders import env_loader, execute_instance_hooks
from dynaconf.loaders.base import BaseLoader
from dynaconf.loaders.yaml_loader import yaml
from dynaconf.utils.files import glob
from dynaconf.utils.functional import empty

from ansible_base.lib.dynamic_config.settings_logic import get_mergeable_dab_settings


def factory(
    module_name: str,  # name of the module that calls this function
    app_name: str,  # main app name to be used to name env_switcher and envvar_prefix
    *,
    add_dab_settings: bool = True,  # add DAB settings to the settings object
    validators: list[Validator] | None = None,  # custom validators to be used in Dynaconf
    extra_envvar_prefixes: list[str] | None = None,  # extra prefixes to be used in envvar loader
    **options,  # options to be passed to Dynaconf
) -> Dynaconf:
    """Create a Dynaconf instance for a specific service.

    This function creates a Dynaconf instance with the following options:

    - env_switcher: based on the app name
    - envvar_prefix: based on the app name
    - root_path: based on the caller module path
    - loaders: empty list (to be loaded manually)
    - options: any other options passed to the function (e.g. settings_files)
      some options are not allowed and will raise an error if passed.

    The Dynaconf instance returned will do the following:

    - Define its mode based on the env_switcher e.g. "AWX_MODE"
    - Define its envvar prefix based on the envvar_prefix e.g. "AWX"
    - Consider multiple envvar_prefixes if extra_envvar_prefixes is passed (e.g. "AWX,FOO,BAR")
    - Define STANDARD_SETTINGS_FILES for /etc/ansible-automation-platform/
      first load yaml files from the base path, then from the app specific subfolder.
    - Pass all the options to the Dynaconf instance (except the invalid ones)
      e.g:
        - `settings_files=["defaults.py"]`
          static defaults for the service
        - `environments=("development", "production", "quiet", "kube")`
          allow the load of additional settings files such as `development_defaults.py`

    All Validator instances passed to the function will be registered in the Dynaconf instance.

    Once the Dynaconf is instantiated (and settings are loaded), the function will
    add the DAB settings to the Dynaconf instance, if  `add_dab_settings` is True,
    and set the `is_development_mode` flag based on the current mode.

    Dynaconf object is returned and the caller can perform additional operations
    such as read settings, set additional settings, load more settings files,
    and finally export the settings to the caller module using the `export` function.
    """

    _check_options(options)
    prefix = app_name.upper()
    app_name = app_name.lower()

    # attempt to get the caller path from the module name first
    # if it fails, fallback to the inspect stack frame back
    if module := sys.modules.get(module_name):
        caller_path = os.path.dirname(inspect.getfile(module))
    else:
        caller_path = os.path.dirname(inspect.getfile(inspect.currentframe().f_back))

    if extra_envvar_prefixes is not None:
        envvar_prefix = ",".join([prefix, *extra_envvar_prefixes])
    else:
        envvar_prefix = prefix

    std_base_path = Path("/etc/ansible-automation-platform/")
    options.setdefault(
        "ANSIBLE_STANDARD_SETTINGS_FILES",
        [
            std_base_path / "settings.yaml",
            std_base_path / "flags.yaml",
            std_base_path / ".secrets.yaml",
            std_base_path / app_name / "settings.yaml",
            std_base_path / app_name / "flags.yaml",
            std_base_path / app_name / ".secrets.yaml",
        ],
    )

    # Set DAB default variables
    options.setdefault(
        "ANSIBLE_BASE_OVERRIDABLE_SETTINGS",
        [
            'INSTALLED_APPS',
            'REST_FRAMEWORK',
            'AUTHENTICATION_BACKENDS',
            'SPECTACULAR_SETTINGS',
            'MIDDLEWARE',
            'OAUTH2_PROVIDER',
            'CACHES',
            'TEMPLATES',
        ],
    )
    options.setdefault("ANSIBLE_BASE_OVERRIDDEN_SETTINGS", [])
    options.setdefault("INSTALLED_APPS", [])
    options.setdefault("MIDDLEWARE", [])
    options.setdefault("REST_FRAMEWORK", {})

    # Create Dynaconf instance with the given options as defaults
    settings = Dynaconf(
        env_switcher=f"{prefix}_MODE",
        envvar_prefix=envvar_prefix,
        root_path=caller_path,
        # Disable default loaders because each file/env will be loaded individually
        loaders=[],
        **options,
    )

    if validators:
        settings.validators.register(*validators)

    add_dab_settings and load_dab_settings(settings)

    # Dynaconf allows composed modes
    # so current_env can be a comma separated string of modes (e.g. "development,quiet")
    settings.set(
        "is_development_mode",
        "development" in settings.current_env.lower(),
        loader_identifier="is_development_mode_check",
    )
    return settings


def export(module_name: str | type, settings: Dynaconf, validation: bool = True):
    """Export the settings from Dynaconf object to the module that called this function.

    module_name is usually `__name__` and is used to get the module object
    arbitrary module objects can be passed as well so testing can be done on
    a fake module object.

    In a django-app/settings.py this will take all variables from the Dynaconf
    instance and set them as attributes in the settings module
    so they can be accessed as `settings.VARIABLE_NAME` from django.conf.settings.

    unless `validation` is set to False, the settings will be validated before
    being exported to the module.
    """
    if isinstance(module_name, str):
        object_to_populate = sys.modules[module_name]
    else:
        object_to_populate = module_name

    validation and validate(settings)
    settings.populate_obj(
        object_to_populate,
        internal=False,
        ignore=["IS_DEVELOPMENT_MODE"],
        convert_to_dict=True,
    )


def load_dab_settings(settings: Dynaconf):
    """Add DAB settings to the settings object."""
    dab_settings = get_mergeable_dab_settings(settings.to_dict())
    settings.set(
        "ANSIBLE_BASE_OVERRIDDEN_SETTINGS",
        list(dab_settings.keys() & settings.ANSIBLE_BASE_OVERRIDABLE_SETTINGS),
        loader_identifier="load_dab_settings",
    )
    settings.update(dab_settings, loader_identifier="load_dab_settings")


def load_standard_settings_files(settings: Dynaconf):
    """Load the standard settings files for Ansible Automation Platform.

    This function loops the `ANSIBLE_STANDARD_SETTINGS_FILES` list in the settings
    and loads each file into the settings object without considering the current env,
    as these files are meant to be environment agnostic.

    NOTE: this function must be replaced by future implementation
    of Dynaconf's `settings.load_file` that will support envless loading.
    """
    loader = BaseLoader(
        obj=settings,
        env=settings.current_env,
        identifier="standard_settings_loader",
        extensions=YAML_EXTENSIONS,
        file_reader=yaml.safe_load,
        string_reader=yaml.safe_load,
        validate=False,
    )
    for path in settings.ANSIBLE_STANDARD_SETTINGS_FILES:
        if not Path(path).exists():
            continue
        data = loader.get_source_data([str(path)])
        loader._envless_load(data)


def load_envvars(settings: Dynaconf):
    """Loads environment variables into the settings object.

    This function exists to be called on-demand after all settings are loaded
    """
    env_loader.load(settings, identifier="dab.dynamic_config")


def _check_options(options):
    """Raise Error if invalid options are passed.

    Invalid options are those that are not allowed in the factory function,
    and must be defined here by Django Ansible Base.
    """
    invalid_options = {
        "envvar_prefix",  # defined based on app name
        "env_switcher",  # defined based on app name
        "root_path",  # defined dynamically by factory
        "validators",  # Cannot be defined because would trigger eager validation
    }
    if invalid_options & set(options.keys()):
        raise ImproperlyConfigured(f"Invalid Dynaconf options: {invalid_options}")


def validate(settings: Dynaconf):
    """Validate the settings according to the validators registered.
    This function operates on a clone of the settings object to avoid
    validation to write to the inspection data.
    """
    if settings.get_environ("BYPASS_DYNACONF_VALIDATION"):
        return

    settings.dynaconf_clone().validators.validate()


def load_python_file_with_injected_context(*paths: str, settings: Dynaconf, run_hooks: bool = True):  # NOSONAR
    """Load a file with context into the settings object.

    This function is not encouraged to be used because it uses `exec` to load,
    the right way to load settings is to use `settings.load_file` method.

    This function exists to keep backwards compatibility with the previous
    usage of `split_settings` that uses `exec` to load the settings files
    with injected context.

    Dynaconf doesn't support this (and will not support it) so this function
    is a temporary solution until all settings files are migrated to be loaded
    by Dynaconf itself from the YAML files only.

    paths: a list of paths to the python files to be loaded, can be absolute or relative or glob
    settings: Dynaconf instance
    """
    # Wanted to raise a warning here but it's not possible to raise a warning without breaking awx CI

    for path in paths:
        pattern = os.path.join(settings.ROOT_PATH_FOR_DYNACONF, path)
        files_to_load = glob(pattern)
        for file_path in files_to_load:
            scope = settings.as_dict()
            file_path = os.path.abspath(file_path)
            with open(file_path, "rb") as to_compile:
                code = compile(to_compile.read(), file_path, "exec")
                exec(code, scope)

            # Update the settings object with the new values coming from the scope
            for key, value in scope.items():
                if key.split("__")[0].isupper():  # this is a setting key
                    # set only if the value have been changed from the executed file
                    if value != settings.get(key, empty):
                        settings.set(
                            key,
                            value,
                            loader_identifier=f"python_file_with_injected_scope:{file_path}",
                        )
                elif key.startswith("_dynaconf_hook") and callable(value):
                    settings._post_hooks.append(value)

            settings._loaded_files.append(file_path)

    if run_hooks:
        execute_instance_hooks(
            settings,
            "post",
            [_hook for _hook in settings._post_hooks if getattr(_hook, "_dynaconf_hook", False) is True and not getattr(_hook, "_called", False)],
        )


def toggle_feature_flags(settings: Dynaconf) -> dict[str, Any]:
    """Toggle FLAGS based on installer settings.
    FLAGS is a django-flags formatted dictionary.
        FLAGS={
            "FEATURE_SOME_PLATFORM_FLAG_ENABLED": [
                {"condition": "boolean", "value": False, "required": True},
                {"condition": "before date", "value": "2022-06-01T12:00Z"},
            ]
        }
    Installers will place `FEATURE_SOME_PLATFORM_FLAG_ENABLED=True/False` in the settings file.
    This function will update the value in the index 0 in FLAGS with the installer value.
    """
    data = {}
    for feature_name, feature_content in settings.get("FLAGS", {}).items():
        if (installer_value := settings.get(feature_name, empty)) is not empty:
            feature_content[0]["value"] = installer_value
            data[f"FLAGS__{feature_name}"] = feature_content
    return data



import logging
import signal
import traceback
import uuid
from inspect import isfunction

from ansible_base.lib.logging import thread_local

logger = logging.getLogger(__name__)


class LogTracebackMiddleware:
    transactions = {}

    @classmethod
    def handle_signal(cls, *args):
        for t_id, request in LogTracebackMiddleware.transactions.items():
            logger.error(
                f"""Received graceful timeout signal for {request.method}
                path: {request.path} while in stack: {''.join(traceback.format_stack())}
                """
            )

    def __init__(self, get_response):
        self.get_response = get_response
        if isfunction(signal.getsignal(signal.SIGABRT)):
            # If signal.getsignal(signal.SIGABRT) returns a function, it means that signal is already being handled.
            raise RuntimeError("SIGABRT is already being handled! Don't use LogTracebackMiddleware!!!")
        try:
            signal.signal(signal.SIGABRT, LogTracebackMiddleware.handle_signal)
        except ValueError:
            logger.error(
                f"""Configured to use {__name__}.LogTracebackMiddleware but the the application
                is not being served in the main thread, so we will not handle SIGABRT."""
            )

    def __call__(self, request):
        t_id = str(uuid.uuid4())
        LogTracebackMiddleware.transactions[t_id] = request
        try:
            return self.get_response(request)
        finally:
            LogTracebackMiddleware.transactions.pop(t_id)


class LogRequestMiddleware:
    """
    Inject the request into the thread local so that it can be accessed by the logging filter.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.process_request(request)
        response = self.get_response(request)
        self.process_response(request, response)
        return response

    def process_request(self, request):
        thread_local.request = request
        return None

    def process_response(self, request, response):
        thread_local.request = None
        return response



# Generated by Django 4.2.6 on 2024-01-18 20:16

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import social_django.storage


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Authenticator',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_on', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('modified_on', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('name', models.CharField(help_text='The name of this resource', max_length=512, unique=True)),
                ('enabled', models.BooleanField(default=False, help_text='Should this authenticator be enabled')),
                ('create_objects', models.BooleanField(default=True, help_text='Allow authenticator to create objects (users, teams, organizations)')),
                ('users_unique', models.BooleanField(default=False, help_text='Are users from this source the same as users from another source with the same id')),
                ('remove_users', models.BooleanField(default=True, help_text='When a user authenticates from this source should they be removed from any other groups they were previously added to')),
                ('configuration', models.JSONField(blank=True, default=dict, help_text='The required configuration for this source')),
                ('type', models.CharField(editable=False, help_text='The type of authentication service this is', max_length=256)),
                ('order', models.IntegerField(default=1, help_text='The order in which an authenticator will be tried. This only pertains to username/password authenticators')),
                ('slug', models.SlugField(default=None, editable=False, help_text='An immutable identifier for the authenticator', max_length=1024, unique=True)),
                ('category', models.CharField(default=None, help_text='The base type of this authenticator', max_length=30)),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
                ('users', models.ManyToManyField(blank=True, help_text='The list of users who have authenticated from this authenticator', related_name='users', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AuthenticatorMap',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_on', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('modified_on', models.DateTimeField(default=None, editable=False, help_text='The date/time this resource was created')),
                ('name', models.CharField(help_text='The name of this resource', max_length=512)),
                ('revoke', models.BooleanField(default=False, help_text='If a user does not meet this rule should we revoke the permission')),
                ('map_type', models.CharField(choices=[('team', 'team'), ('is_superuser', 'is_superuser'), ('is_system_auditor', 'is_system_auditor'), ('allow', 'allow'), ('organization', 'organization')], default='team', help_text='What does the map work on, a team, a user flag or is this an allow rule', max_length=17)),
                ('team', models.CharField(blank=True, default=None, help_text='A team name this rule works on', max_length=512, null=True)),
                ('organization', models.CharField(blank=True, default=None, help_text='An organization name this rule works on', max_length=512, null=True)),
                ('triggers', models.JSONField(blank=True, default=dict, help_text='Trigger information for this rule')),
                ('order', models.PositiveIntegerField(default=0, help_text='The order in which this rule should be processed, smaller numbers are of higher precedence. Items with the same order will be executed in random order')),
                ('authenticator', models.ForeignKey(help_text='The authenticator this mapping belongs to', on_delete=django.db.models.deletion.CASCADE, to='dab_authentication.authenticator')),
                ('created_by', models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL)),
                ('modified_by', models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='AuthenticatorUser',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(db_index=True, max_length=255)),
                ('extra_data', models.JSONField(default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('claims', models.JSONField(blank=True, default=dict)),
                ('last_login_map_results', models.JSONField(blank=True, default=list)),
                ('access_allowed', models.BooleanField(default=None, null=True)),
                ('provider', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='authenticator_user', to='dab_authentication.authenticator', to_field='slug')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='authenticator_user', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('provider', 'uid')},
            },
            bases=(models.Model, social_django.storage.DjangoUserMixin),
        ),
        migrations.AddConstraint(
            model_name='authenticatormap',
            constraint=models.CheckConstraint(check=models.Q(models.Q(('map_type', 'team'), _negated=True), models.Q(('team__isnull', False), ('organization__isnull', False)), _connector='OR'), name='dab_authentication_authenticatormap_require_org_team_if_team_map'),
        ),
        migrations.AddConstraint(
            model_name='authenticatormap',
            constraint=models.CheckConstraint(check=models.Q(models.Q(('map_type', 'organization'), _negated=True), ('organization__isnull', False), _connector='OR'), name='dab_authentication_authenticatormap_require_org_if_org_map'),
        ),
        migrations.AlterUniqueTogether(
            name='authenticatormap',
            unique_together={('name', 'authenticator')},
        ),
    ]



from django.contrib.auth import get_user_model
from rest_framework.decorators import action
from rest_framework.response import Response

from ansible_base.lib.abstract_models.common import get_cls_view_basename
from ansible_base.lib.utils.response import get_relative_url
from ansible_base.oauth2_provider.models import OAuth2AccessToken
from ansible_base.oauth2_provider.serializers import OAuth2TokenSerializer


class DABOAuth2UserViewsetMixin:
    """
    This mixin provides several actions to expose as sub-urls for a given user PK.
    """

    def extra_related_fields(self, obj) -> dict[str, str]:
        fields = super().extra_related_fields(obj)
        user_basename = get_cls_view_basename(get_user_model())
        fields['personal_tokens'] = get_relative_url(f'{user_basename}-personal-tokens-list', kwargs={"pk": obj.pk})
        fields['authorized_tokens'] = get_relative_url(f'{user_basename}-authorized-tokens-list', kwargs={"pk": obj.pk})
        fields['tokens'] = get_relative_url(f'{user_basename}-tokens-list', kwargs={"pk": obj.pk})
        return fields

    def _user_token_response(self, request, filters, pk):
        tokens = OAuth2AccessToken.objects.filter(user=pk, **filters)
        page = self.paginate_queryset(tokens)
        if page is not None:
            serializer = OAuth2TokenSerializer(page, many=True, context={"request": request})
            return self.get_paginated_response(serializer.data)

        serializer = OAuth2TokenSerializer(tokens, many=True)
        return Response(serializer.data)

    @action(detail=True, methods=["get"], url_name="personal-tokens-list")
    def personal_tokens(self, request, pk=None):
        filters = {'application__isnull': True}
        return self._user_token_response(request, filters, pk)

    @action(detail=True, methods=["get"], url_name="authorized-tokens-list")
    def authorized_tokens(self, request, pk=None):
        filters = {'application__isnull': False}
        return self._user_token_response(request, filters, pk)

    @action(detail=True, methods=["get"], url_name="tokens-list")
    def tokens(self, request, pk=None):
        return self._user_token_response(request, {}, pk)



import logging

from channels.auth import AuthMiddleware
from channels.auth import get_user as get_session_user
from channels.db import database_sync_to_async
from channels.security.websocket import WebsocketDenier
from channels.sessions import CookieMiddleware, SessionMiddleware
from django.contrib.auth import get_user_model
from django.http import HttpRequest
from rest_framework.request import Request
from rest_framework.settings import api_settings

logger = logging.getLogger('ansible_base.lib.channels.middleware')


@database_sync_to_async
def _get_authenticated_user(scope: dict):
    request = HttpRequest()
    request.META = {_http_key(k.decode()): v.decode() for (k, v) in scope["headers"]}
    auth_classes = [auth() for auth in api_settings.DEFAULT_AUTHENTICATION_CLASSES]
    try:
        return Request(request, authenticators=auth_classes).user
    except Exception:
        return None


class DrfAuthMiddleware(AuthMiddleware):
    async def __call__(self, scope, receive, send):
        session_user = await get_session_user(scope)
        if session_user and session_user.is_authenticated:
            user = session_user
        else:
            user = await _get_authenticated_user(scope)

        if not user or not isinstance(user, get_user_model()):
            logger.error("Websocket connection does not provide valid authentication")
            denier = WebsocketDenier()
            return await denier(scope, receive, send)

        scope["user"] = user

        return await self.inner(scope, receive, send)


# Handy shortcut for applying all three layers at once
def DrfAuthMiddlewareStack(inner):  # noqa: N802
    return CookieMiddleware(SessionMiddleware(DrfAuthMiddleware(inner)))


def _http_key(key: str) -> str:
    return f"HTTP_{key.replace('-', '_').upper()}"
